"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_cents.py $
$Id: utest_cents.py 27635 2005-10-28 19:43:18Z dbinger $
"""
from dulcinea.cents import Cents
from qp.lib.spec import get_spec_problems
from sancho.utest import UTest


class CentsTest(UTest):

    def notice_bad(self):
        try:
            a = Cents(None)
            assert 0
        except TypeError: pass
        a = Cents(0)
        assert get_spec_problems(a) == []
        try:
            a + 1
            assert 0
        except TypeError: pass
        try:
            a * 1
            assert 0
        except TypeError: pass
        try:
            a / 1
            assert 0
        except TypeError: pass
        try:
            a > 1
            assert 0
        except TypeError: pass
        try:
            a += 1
            assert 0
        except TypeError: pass
        try:
            a -= 1
            assert 0
        except TypeError: pass

    def do_good_things(self):
        a = Cents(23)
        assert 23 == a.get_cents()
        assert 0.23 == a.get_dollars(), a.get_dollars()
        assert str(a) == '$0.23'
        assert (a - a).get_cents() == 0
        assert (a + a).get_cents() == 46
        assert a == Cents(23)
        a += Cents(5)
        assert a == Cents(28)
        a -= Cents(30)
        assert a == Cents(-2)

if __name__ == '__main__':
    CentsTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/survey.py $
$Id: survey.py 30484 2008-01-21 20:41:06Z dbinger $
"""
from dulcinea.address import ContactAddress
from dulcinea.base import DulcineaPersistent, Copyable
from dulcinea.timestamped import Timestamped
from dulcinea.user import DulcineaUser
from durus.persistent_dict import PersistentDict
from durus.persistent_list import PersistentList
from qp.lib.keep import Keyed, Keep
from qp.lib.spec import either, sequence, mapping, require, init, specify
from qp.lib.spec import spec, boolean, add_getters_and_setters, add_getters
from qp.lib.spec import string, proper, Specified, Mixin
from qp.pub.common import get_publisher

def get_question_db():
    return get_publisher().get_connection().get_root().get('question_db')

def get_surveys_and_questions_db():
    return get_publisher().get_connection().get_root().get(
        'surveys_and_questions_db')

class Question (DulcineaPersistent, Keyed):

    title_is = (string, None)
    sub_title_is = (string, None)

    def __init__(self, title, sub_title=None):
        assert self.__class__ is not Question, "abstract class"
        Keyed.__init__(self)
        specify(self, title=title, sub_title=sub_title)

    def __str__(self):
        if self.sub_title:
            return self.title + "\n" + self.sub_title
        return self.title

    def get_new_answer(self):
        raise NotImplementedError

add_getters(Question)


class TextQuestion (Question):

    dimensions_is = (int, int)

    def __init__(self, title, sub_title=None, rows=5, cols=80):
        Question.__init__(self, title, sub_title)
        self.dimensions = (rows, cols)

    def get_new_answer(self):
        return TextAnswer(self)

    def get_dimensions(self):
        # Fix with update_db.
        return getattr(self, 'dimensions', (5, 80))


class ContactAddressQuestion (Question):

    def get_new_answer(self):
        return ContactAddressAnswer(self)


class SelectQuestion (Question):
    """
    A SelectQuestion is an abstract class for representing selections
    from either checkboxes or radio buttons
    """
    column_headings_is = spec(
        [(string, None)],
        "text to place above each column")
    row_headings_is = spec(
        [(string, None)],
        "text that preceeds each row")
    range_labels_is = spec(
        [string],
        "labels to be placed in a row above column headings at the right and "
        "left")
    last_row_has_text_is = spec(
        boolean,
        "flag to tell the widget to display a TextWidget after the last row "
        "label")

    def __init__(self, title, sub_title=None,
                 column_headings=None, row_headings=None,
                 range_labels=None, last_row_has_text=False):
        Question.__init__(self, title, sub_title=sub_title)
        specify(self,
                column_headings=column_headings or [],
                row_headings=row_headings or [],
                range_labels=range_labels or [],
                last_row_has_text = last_row_has_text)
        assert len(self.range_labels) in (0,2), "must have two range labels"

    def __str__(self):
        question = self.title
        if len(self.column_headings) > 1: # Radioset
            if len(self.column_headings) > 4:
                headings = '(%s ... %s)' % (self.column_headings[0],
                                            self.column_headings[-1])
            else:
                headings = '(%s)' % (', '.join(self.column_headings))
            question += "\n" + headings
            if self.range_labels:
                question += "\n  ...where %s is %s and %s is %s" % (
                    self.column_headings[0], self.range_labels[0],
                    self.column_headings[-1], self.range_labels[-1])
        if self.sub_title:
            question += "\n" + self.sub_title
        return question

    def get_new_answer(self):
        return SelectAnswer(self)

add_getters(SelectQuestion)


class Answer (Specified, Copyable):

    question_is = Question

    def __init__(self, question=None):
        assert self.__class__ is not Answer, "abstract class"
        self.question = question

    def get_question(self):
        return self.question


class ContactAddressAnswer (Answer):

    answer_is = ContactAddress

    def __init__(self, question=None):
        Answer.__init__(self, question)
        self.answer = ContactAddress()

    def __str__(self):
        return self.answer.format()

    def has_answer(self):
        return self.answer != ContactAddress()

add_getters_and_setters(ContactAddressAnswer)


class TextAnswer (Answer):

    answer_is = (None, string)

    def __init__(self, question=None):
        Answer.__init__(self, question)
        self.answer = None

    def __str__(self):
        return self.answer

    def get_answer(self):
        return self.answer

    def set_answer(self, answer):
        specify(self, answer=answer)

    def has_answer(self):
        return self.answer is not None

add_getters_and_setters(TextAnswer)


class SelectAnswer (Answer):

    selected_is = spec(
        (None, [either(string, boolean)]),
        "list of selected row answers")
    other_text_is = spec(
        (None, string),
        "contents of the filled in 'last_row_has_text' textfield")

    def __init__(self, question=None):
        Answer.__init__(self, question)
        init(self)

    def __str__(self):
        if self.question.row_headings:
            def format_heading_answer(heading, answer):
                if answer:
                    return "%s: %s" % (heading, answer)
                return ''
            return "\n".join([format_heading_answer(*heading_answer) for
                              heading_answer in
                              zip(self.question.row_headings, self.selected)])
        return ", ".join(self.selected)

    def set_other_text(self, other_text):
        specify(self, other_text=other_text)

    def get_other_text(self):
        return self.other_text

    def set_answer(self, selected):
        specify(self, selected=selected)

    def get_answer(self):
        return self.selected

    def has_answer(self):
        return self.selected is not None

add_getters(SelectAnswer)


class Survey (DulcineaPersistent, Timestamped):
    """A set of Questions and Answers attached to a single object.
    """
    user_is = spec(
        (DulcineaUser, None),
        "user taking the survey")
    answers_is = spec(
        [proper(Answer)],
        "answers to the questions that comprise the object's survey")

    def __init__(self, answers):
        """(answers:[Answer])
        """
        Timestamped.__init__(self)
        init(self, answers=answers)

    def __str__(self):
        return "By %s on %s: %d answers" % (
            self.user, self.get_timestamp(), len(self.answers))

    def __len__(self):
        return len(self.answers)

    def get_questions(self):
        return [answer.get_question() for answer in self.answers]

    def get_answer_for_question_key(self, question_key):
        for answer in self.answers:
            if answer.get_question().get_key() == question_key:
                return answer
        return None

    def set_user(self, user):
        """(user : DulcineaUser)
        """
        specify(self, user=user)

    def set_answers(self, answers, user):
        """(answers : [Answer], user : DulcineaUser)
        """
        specify(self, answers=answers, user=user)
        self.set_timestamp()

    def has_answers(self, omit_questions=None):
        """(omit_questions : [int] | None) -> boolean

        Return true if any answers have been answered for this survey
        ignoring omitted questions if specified
        """
        for answer in self.answers:
            if answer.get_question().get_key() in (omit_questions or []):
                continue
            if answer.has_answer():
                return True
        return False

add_getters(Survey)


class Surveyable (Mixin):
    """A mixin for objects that have Surveys
    """
    surveys_is = (None, sequence(Survey, PersistentList))

    def __init__(self):
        self.surveys = None

    def get_surveys(self):
        """() -> [Survey]
        """
        return self.surveys or []

    def get_survey(self, index):
        return self.surveys and self.surveys[index]

    def get_new_survey(self):
        """() -> Survey
        """
        raise NotImplementedError

    def add_survey(self, survey):
        if self.surveys is None:
            self.surveys = PersistentList()
        self.surveys.append(survey)


class QuestionDatabase (DulcineaPersistent, Keep):

    def __init__(self):
        Keep.__init__(self, value_spec=Question)

    def get_answers(self, question_keys=None):
        """(question_keys: [int]) -> [Answer]
        """
        if not question_keys:
            question_keys = sorted(self.get_mapping().keys())
        return [self.get(question_key).get_new_answer()
                for question_key in question_keys]


class SurveysAndQuestions (QuestionDatabase, Surveyable):

    id_is = string
    title_is = string
    description_is = string

    def __init__(self, id, questions):
        """(id : string, questions : [Question])
        """
        Surveyable.__init__(self)
        QuestionDatabase.__init__(self)
        for question in questions:
            self.add(question)
        specify(self, id=id, title='', description='')

    def get_new_survey(self):
        return Survey(self.get_answers())

add_getters_and_setters(SurveysAndQuestions)


class SurveysAndQuestionsDatabase (DulcineaPersistent):

    surveys_and_questions_is = mapping({string:SurveysAndQuestions},
                                       PersistentDict)

    def __init__(self):
        self.surveys_and_questions = PersistentDict()

    def add(self, surveys_and_questions):
        require(surveys_and_questions, SurveysAndQuestions)
        id = surveys_and_questions.get_id()
        assert id not in self.surveys_and_questions, id
        self.surveys_and_questions[id] = surveys_and_questions

    def get(self, id):
        return self.surveys_and_questions.get(id)

    def get_mapping(self):
        return self.surveys_and_questions




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/attachable.py $
$Id: attachable.py 29832 2007-05-29 19:19:08Z dbinger $

Provides the Attachable mixin class that allows objects to to have
file attachments.
"""
from dulcinea.stored_file import StoredFile, MIME_TYPES
from dulcinea.user import DulcineaUser
from durus.persistent import PersistentObject
from qp.lib.spec import spec, specify, add_getters_and_setters, require, string
from qp.lib.spec import Specified, Mixin, datetime_with_tz
from qp.pub.common import site_now


class Attachment (PersistentObject, Specified):
    """
    Stores properties of an attachment relationship between an Attachable
    object and a StoredFile.
    """
    file_is = spec(
        StoredFile,
        "the file attached")
    owner_is = spec(
        DulcineaUser,
        "the owner of this association")
    date_is = spec(
        datetime_with_tz,
        "date this attachment was created")
    filename_is = spec(
        string,
        "the filename to use when downloading the file")
    description_is = spec(
        (None, string),
        "a description of the file")

    def __init__(self, file, owner):
        specify(self,
                file=file,
                owner=owner,
                date=site_now(),
                filename=file.get_filename(),
                description=file.get_description())

    def open(self):
        return self.file.open()

    def get_file_id(self):
        return self.file.get_id()

    def get_size(self):
        return self.file.get_size()

    def get_mime_type(self):
        return self.file.get_mime_type()

    def has_manage_access(self, user):
        return (user and (user is self.owner or user.is_admin()))

    def is_image(self):
        return self.get_mime_type().startswith('image/')

add_getters_and_setters(Attachment)


class Attachable (Mixin):

    attachments_is = spec(
        [Attachment],
        "attachments connected to this object")

    def __init__(self, attachments=None):
        specify(self, attachments=list(attachments or []))

    def clear_attachments(self):
        specify(self, attachments=[])

    def get_attachments(self):
        """() -> [Attachment]
        Retrieve a list of the attachments for this object.
        """
        return self.attachments

    def get_attached_files(self):
        """() -> [StoredFile]
        Retrieve a list of the file objects attached to this object.
        """
        return [a.get_file() for a in self.attachments]

    def add_attachment(self, attachment):
        require(attachment, Attachment)
        for existing_attachment in self.attachments:
            if existing_attachment.get_file_id() == attachment.get_file_id():
                return
        self.attachments = self.attachments + [attachment]

    def get_attachment(self, id):
        """(id : string) -> Attachment | None
        """
        for attachment in self.attachments:
            if attachment.get_file_id() == id:
                return attachment
        return None

    def attach_file(self, stored_file, user):
        """(stored_file: StoredFile)
        Attach a file to this object.
        """
        files = self.attachments[:]
        files.append(Attachment(stored_file, user))
        self.attachments = files

    def detach_attachment(self, attachment):
        """(attachment: Attachment)
        Detach an attachment from this object.
        """
        require(attachment, Attachment)
        files = self.attachments[:]
        files.remove(attachment)
        self.attachments = files

    def attachment_modified(self, attachment, user):
        """(attachment : Attachment, user : DulcineaUser)

        'attachment' was modified by 'user'.
        Override this method to intercept modification activity for attachments
        """
        pass

    def get_allowed_mime_types(self, user=None):
        """() -> [string]

        Override to restrict the types of files that can be attached
        """
        return [mime_type[0] for mime_type in MIME_TYPES]




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_calendar.py $
$Id: utest_calendar.py 31118 2008-09-16 14:55:40Z dbinger $
"""
from datetime import datetime, timedelta
from dulcinea.calendar import Calendar, Reservation, Reservable, Resources
from dulcinea.calendar import Resource, weekdays
from dulcinea.user import DulcineaUser
from sancho.utest import UTest, raises
from qp.lib.tz import Eastern
from dulcinea.base import DulcineaPersistent

class PersistentReservable(Reservable, DulcineaPersistent):
    pass


class Test (UTest):

    def check_a(self):
        calendar = Calendar()
        raises(TypeError, calendar.add, Reservation())
        block1 = Reservation()
        title = 'test1'
        block1.set_title(title)
        assert block1.get_title() == title
        block1.set_description(title)
        assert block1.get_description() == title
        reserved = Resource()
        block1.set_reserved(reserved)
        assert block1.get_reserved() == reserved
        user = DulcineaUser(None)
        block1.set_reserver(user)
        start1 = datetime(1,1,1, tzinfo=Eastern)
        block1.set_start_time(start1)
        assert block1.get_start_time() == start1
        end1 = start1 + timedelta(days=1)
        block1.set_end_time(end1)
        assert block1.get_end_time() == end1
        calendar.add(block1)
        assert calendar.get(block1.get_key()) == block1
        assert calendar.get(-1, None) == None
        assert dict(calendar.get_mapping()) == { block1.get_key():  block1 }
        for entry in calendar:
            assert entry is block1
        reservable = PersistentReservable()
        reservable.set_active(True)
        assert reservable.is_active() == True
        title = 'room 1'
        reservable.set_title(title)
        assert reservable.get_title() == title
        block1.set_reserved(reservable)
        assert reservable.allows_control(user) == False
        assert reservable.allows_reserve(user) == False
        class AdminUser(DulcineaUser):
            def is_admin(self):
                return True
        admin_user = AdminUser('me')
        assert block1.can_modify(None) == False
        assert block1.can_modify(admin_user) == True
        block1.set_reserver(admin_user)
        assert block1.can_modify(user) == False
        user.get_permissions().grant('control', reservable)
        assert block1.can_modify(user) == True
        user2 = DulcineaUser('u')
        block1.set_reserver(user2)
        assert block1.can_modify(user2) == True
        Resources()
        time_increment = timedelta(hours=2)
        start_time = start1
        end_time = start_time + timedelta(days=3)
        start_time_of_day = timedelta(hours=8)
        end_time_of_day = timedelta(hours=18)
        resource = Resource()
        reservation = Reservation()
        reservation.set_title("")
        reservation.set_description("")
        reservation.set_reserved(resource)
        reservation.set_reserver(user)
        reservation.set_start_time(start_time + start_time_of_day)
        reservation.set_end_time(start_time + start_time_of_day + 
                                 2 * time_increment)
        reservation.set_reserved(resource)
        calendar.add(reservation)
        calendar.get_table_map(
            resource, time_increment, start_time, end_time, 
            start_time_of_day, end_time_of_day)

    def check_weekdays(self):
        failures = False
        for length in range(9):
            for offset in range(9):
                start = datetime(
                    2000, 1, 1, tzinfo=Eastern) + timedelta(days=offset)
                end = start + timedelta(days=length)
                weekdays1 = weekdays(start, end)
                currentday = start
                weekdays2 = 0
                for i in range(length):
                    if 5 <= currentday.weekday() <= 6:
                        pass
                    else:
                        weekdays2 += 1
                    currentday += timedelta(days=1)
                if weekdays1 != weekdays2:
                    failures = True
        assert not failures

if __name__ == "__main__":
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_set_utils.py $
$Id: utest_set_utils.py 30177 2007-10-19 11:30:16Z dbinger $
"""
from dulcinea.physical_unit import  UnitCollection
from dulcinea.physical_value import PhysicalValue
from dulcinea.range_value import RangeValue
from dulcinea.set_utils import _get_physical_value_intersection
from dulcinea.set_utils import get_set_intersection
from dulcinea.set_utils import in_set, get_range_intersection
from sancho.utest import UTest, raises

def split_comma (s):
    return s.split(', ')

class CompareTest (UTest):

    def _pre(self):
        self.units = units = UnitCollection()

    def _post(self):
        del self.units

    def check_simple_set(self):
        set1 = ['foo', 'bar']
        assert in_set('foo', set1)
        assert not in_set('', set1)
        assert not in_set(3, set1)
        set1.append(3)
        assert in_set(3, set1)
        raises(TypeError, in_set, 'foo', 'bar')
        raises(TypeError, in_set, 'f', 'bar')
        raises(TypeError, in_set, 'f', 3)

    def check_range_set(self):
        set2 = [3, 6, RangeValue(8, 10)]
        assert in_set(3, set2)
        assert not in_set(4, set2)
        assert not in_set('foo', set2)
        assert in_set(6.0, set2)
        assert in_set(6, set2)
        assert in_set(8, set2)
        assert in_set(9, set2)
        assert in_set(10, set2)
        assert in_set(8.5, set2)
        assert not in_set(10.5, set2)

        # Put the range in the middle
        set2.append(12)
        assert in_set(10, set2)
        assert not in_set(10.5, set2)
        assert not in_set(11, set2)
        assert in_set(12, set2)

    def check_pv_set(self):
        get_unit = self.units.get_unit
        set3 = (PhysicalValue(3, get_unit("min")),
                PhysicalValue(4, get_unit("s")),
                PhysicalValue(1.5, get_unit("hour")))
        assert in_set(PhysicalValue(3, get_unit('min')), set3)
        assert not in_set(PhysicalValue(3.1, get_unit('min')), set3)
        assert in_set(PhysicalValue(180, get_unit('s')), set3)
        assert in_set(PhysicalValue(1.5, get_unit('hour')), set3)
        assert in_set(PhysicalValue(90, get_unit('min')), set3)
        assert not in_set(PhysicalValue(-90, get_unit('min')), set3)

    def check_pv_range_set(self):
        get_unit = self.units.get_unit
        set4 = (PhysicalValue(1.5, get_unit("m")),
                PhysicalValue(RangeValue(200, 300), get_unit("cm")),
                PhysicalValue(4, get_unit("m")))
        assert in_set(PhysicalValue(1.5, get_unit('m')), set4)
        assert not in_set(PhysicalValue(1.4, get_unit('m')), set4)
        assert in_set(PhysicalValue(2, get_unit('m')), set4)
        assert in_set(PhysicalValue(2.9, get_unit('m')), set4)
        assert in_set(PhysicalValue(3000, get_unit('mm')), set4)
        assert in_set(PhysicalValue(2.5e6, get_unit('um')), set4)
        assert in_set(PhysicalValue(4e6, get_unit('um')),
                      [PhysicalValue(4, get_unit('m'))])
        assert in_set(PhysicalValue(4e6, get_unit('um')), set4)
        assert not in_set(PhysicalValue(1, get_unit('s')), set4)

def _parse_physical_value(s, get_unit):
    split = s.split()
    if len(split) == 1:
        return PhysicalValue(float(split[0]), None)
    elif len(split) == 2:
        return PhysicalValue(float(split[0]), get_unit(split[1]))
    else:
        return PhysicalValue(RangeValue(float(split[0]),
                                        float(split[2])),
                             get_unit(split[3]))

class IntersectionTest (UTest):

    def _pre(self):
        self.units = units = UnitCollection()

    def _post(self):
        del self.units

    def _testrv (self, a, b, expect):
        ra = RangeValue(*a)
        rb = RangeValue(*b)
        assert get_range_intersection(ra, rb) == RangeValue(*expect)
        assert get_range_intersection(rb, ra) == RangeValue(*expect)

    def _testp (self, a, b, expect):
        pa = _parse_physical_value(a, self.units.get_unit)
        pb = _parse_physical_value(b, self.units.get_unit)
        pe = expect and _parse_physical_value(expect, self.units.get_unit)
        assert _get_physical_value_intersection(pa, pb) == pe
        assert _get_physical_value_intersection(pb, pa) == pe

    def _tests (self, a, b, expect):
        pa = [_parse_physical_value(x, self.units.get_unit)
              for x in split_comma(a)]
        pb = [_parse_physical_value(x, self.units.get_unit)
              for x in split_comma(b)]
        pe = expect and [_parse_physical_value(x, self.units.get_unit)
                         for x in  split_comma(expect)]
        assert get_set_intersection(pa, pb) == pe
        assert get_set_intersection(pb, pa) == pe

    def check_intersections_of_range_value(self):
        self._testrv((0.0, 1.0), (0.0, 0.9), (0.0, 0.90))
        self._testrv((0.0, 1.0), (0.0, 1.0), (0.0, 1.0))
        self._testrv((0.0, 1.0), (0.0, 1.1), (0.0, 1.0))
        self._testrv((0.0, 1.0), (0.1, 0.9), (0.1, 0.90))
        self._testrv((0.0, 1.0), (0.1, 1.0), (0.1, 1.0))
        self._testrv((0.0, 1.0), (0.1, 1.1), (0.1, 1.0))
        self._testrv((0.0, 1.0), (-0.1, 0.9), (0.0, 0.90))
        self._testrv((0.0, 1.0), (-0.1, 1.0), (0.0, 1.0))
        self._testrv((0.0, 1.0), (-0.1, 1.1), (0.0, 1.0))

    def check_intersections_of_physical_value(self):
        self._testp("0 um", "0 .. 5 um", "0 um")
        self._testp("-1 um", "0 .. 5 um", None)
        self._testp("1 um", "0 .. 5 um", "1 um")
        self._testp("1 .. 3 um", "0 .. 5 um", "1 .. 3 um")
        self._testp("1 .. 6 um", "0 .. 5 um", "1 .. 5 um")
        self._testp("1 .. 6 um", "9 .. 11 um", None)

    def check_intersections_of_lists_of_physical_value(self):
        self._tests("1 um, 3 um", "3 um, 5 um", "3 um")
        self._tests("1 um, 3 um", "3, 5 um", [])
        self._tests("1 um, 3 um", "3 .. 5 um", "3 um")
        self._tests("1 um, 3 um, 5 um", "3 .. 5 um", "3 um, 5 um")
        self._tests("1 um, 3 .. 5 um", "3 um, 5 um", "3 um, 5 um")
        self._tests("1 um, 3 um", "1 um, 5 um", "1 um")

if __name__ == "__main__":
    CompareTest()
    IntersectionTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/partition.py $
$Id: partition.py 31356 2008-11-25 16:37:59Z dbinger $
"""

from durus.btree import BTree
from durus.persistent import PersistentObject
from qp.lib.keep import Keyed, Counter
from qp.lib.spec import require, get_spec_problems, anything, mapping, spec
from qp.lib.spec import add_getters, Specified, Mixin

class PartKeyed (Keyed, Mixin):

    part_key_is = anything

    def __init__(self):
        assert isinstance(self, PersistentObject)
        Keyed.__init__(self)
        self.part_key = None

    def get_part_key(self):
        return self.part_key

class Partition (PersistentObject, Specified):

    value_spec_is = spec(
        anything,
        "Specifies the type of values allowed in mappings")
    part_key_spec_is = spec(
        anything,
        "Specifies the type of keys allowed in the partition mapping")
    parts_is = mapping({anything:BTree}, BTree)
    key_counter_is = Counter

    def __init__(self, value_spec=PartKeyed, part_key_spec=anything):
        self.parts = BTree()
        self.key_counter = Counter()
        self.value_spec = value_spec
        self.part_key_spec = part_key_spec

    def get(self, key, default=None):
        for part in self.parts.itervalues():
            value = part.get(key, None)
            if value is not None:
                return value
        return default

    def __nonzero__(self):
        return self.parts.__nonzero__()
    __bool__ = __nonzero__

    def get_max_item(self):
        """() -> (key:anything, value:anything)
        Return the part item with the maximal key.
        """
        maximal = (0, None)
        for part in self.parts.iteritems():
            part_max_item = part[1].get_max_item()
            if part_max_item[0] > maximal[0]:
                maximal = part_max_item
        assert maximal != (0, None), "empty Partition has no max item"
        return maximal

    def itervalues(self):
        for part in self.parts.itervalues():
            for value in part.itervalues():
                yield value

    def iterkeys(self):
        for part in self.parts.itervalues():
            for key in part.iterkeys():
                yield key

    def iteritems(self):
        for part in self.parts.itervalues():
            for item in part.iteritems():
                yield item

    def get_part(self, part_key, default=None):
        return self.parts.get(part_key, default)

    def remove(self, value):
        part = self.parts[value.part_key]
        del part[value.key]
        if not part:
            del self.parts[value.part_key]

    def discard(self, value):
        if value.part_key is not None:
            part = self.parts.get(value.part_key)
            if part and value.key in part:
                self.remove(value)

    def move(self, value, part_key):
        if value.get_part_key() != part_key:
            require(part_key, self.part_key_spec)
            self.discard(value)
            assert self.get(value.key) is None
            value.part_key = part_key
            if get_spec_problems(value):
                raise TypeError(''.join(get_spec_problems(value)))
            if part_key not in self.parts:
                self.parts[part_key] = BTree()
            self.parts[part_key][value.key] = value

    def add(self, value, part_key):
        require(value, self.value_spec)
        assert value.key is None
        assert value.part_key is None
        value.key = self.key_counter.next()
        self.move(value, part_key)

add_getters(Partition)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/history.py $
$Id: history.py 30268 2007-11-13 19:02:20Z dbinger $
"""
from dulcinea.user import DulcineaUser
from durus.persistent import PersistentObject
from durus.utils import xrange
from qp.lib.spec import add_getters, specify, require, string, proper, Mixin
from qp.lib.spec import Specified, datetime_with_tz
from qp.pub.common import get_users, site_now

class Event (Specified):
    """
    Event objects should never be created outside of History.add_event().
    """
    timestamp_is = datetime_with_tz
    user_is = (DulcineaUser, None)
    event_code_is = string
    message_is = (string, None)

    def __init__(self, *args):
        if args:
            user, event_code, message = args
            specify(self, user=user, event_code=event_code, message=message,
                    timestamp=site_now())

    def __str__(self):
        return "%s %s" % (self.timestamp, self.event_code)

    def get_user(self):
        if self.user is None:
            return get_users().get('')
        return self.user

add_getters(Event)


class History (PersistentObject, Specified):

    events_is = [proper(Event)]

    def __init__(self):
        specify(self, events=[])

    def __str__(self):
        return "history (%d events)" % len(self.events)

    def __len__(self):
        return len(self.events)

    def __getitem__(self, i):
        return self.events[i]

    def add_event(self, user, event_code, message):
        """(user : DulcineaUser, event_code : string, message : string)

        Create and append a new Event instance. 
        """
        new_event = Event(user, event_code, message)
        self._p_note_change()
        self.events.append(new_event)

    def update_last_event(self, user, event_code, message):
        """(user : DulcineaUser, event_code : string, message : string)

        Create a new Event instance. If the last event on the list had
        the same event code and user, replace the last event with the
        new one.
        """
        self._p_note_change()
        if self.events:
            if (self.events[-1].event_code == event_code and
                self.events[-1].user == user):
                self.events = self.events[:-1]
        self.events.append(Event(user, event_code, message))

    def find_event(self, user=None, event_code=None, event_codes=None,
                   latest=False, message=None):
        """(user : DulcineaUser = None,
            event_code : string = None,
            latest : bool = False,
            message : string = None) -> Event

        Find the earliest(or latest, if latest is true) event
        matching 'user' and/or 'event_code'.  If either search criterion
        is not supplied or None, it does not affect the search, thus the
        default behaviour is to return the first event in the history
        list.  Return None if no events match.
        """
        if user is not None:
            require(user, DulcineaUser)
        if latest:
            indices = xrange(len(self.events)-1, -1, -1)
        else:
            indices = xrange(len(self.events))
        for i in indices:
            event = self.events[i]
            if user and event.user is not user:
                continue
            if event_code is not None and event.event_code != event_code:
                continue
            if event_codes is not None and event.event_code not in event_codes:
                continue
            if message is not None and event.message != message:
                continue
            return event
        else:
            return None

    def find_events_after(self, time, event_code=None):
        return [event
                for event in self.events
                if event.get_timestamp() > time and (
                    event_code is None or
                    event.get_event_code() == event_code)]

    def find_last_date(self, event_code):
        """(event_code:string) -> datetime | None"""
        event = self.find_event(event_code=event_code, latest=True)
        return event and event.get_timestamp()


class Historical (Mixin):

    history_is = History

    def __init__(self):
        self.history = History()

    def get_history(self):
        return self.history



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_numeric.py $
$Id: utest_numeric.py 27147 2005-08-03 13:59:08Z dbinger $
"""
from dulcinea.numeric import near_cmp
from sancho.utest import UTest

class CompareTest (UTest):

    def check_nonfloat (self):
        assert near_cmp(1, 2) == -1
        assert near_cmp(2, 1) == +1
        assert near_cmp(2, 2) == 0

        assert near_cmp('foo', 'bar') > 0
        assert near_cmp('bar', 'foo') < 0
        assert not near_cmp('foo', 'foo')

    def check_float (self):
        assert near_cmp(1.0, 1.0) == 0
        assert near_cmp(1.0, 1.00001) == -1
        assert near_cmp(1.000001, 1.0) == +1

        assert near_cmp(1.0e-9, 1.0e-9) == 0
        assert near_cmp(1.0e-9, 1.00001e-9) == -1
        assert near_cmp(1.000001e-9, 1.0e-9) == +1

        assert near_cmp(10.0, -10.0) == +1
        assert near_cmp(-10.0, 10.0) == -1
        assert near_cmp(-10.0, -10.0) == 0
        assert near_cmp(-10.0, -20.0) == +1
        assert near_cmp(-20.0, -10.0) == -1

        # provoke some round-off error
        v1 = 1000.0
        v2 = 1.0e-3 / 1.0e-6
        assert abs(v1 - v2) > 0
        assert v1 != v2
        assert near_cmp(v1, v2) == 0
        assert near_cmp(v2, v1) == 0

    def check_zero (self):
        small = 1e-6
        very_small = 1e-18

        assert near_cmp(0.0, 0.0) == 0
        assert near_cmp(0.0, small) == -1
        assert near_cmp(0.0,-small) == +1
        assert near_cmp(0.0, very_small) == 0
        assert near_cmp(0.0, -very_small) == 0

    def check_mixed (self):
        assert near_cmp(1, 1.0) == 0
        assert near_cmp(1.0, 1) == 0
        assert near_cmp(1, 1.00001) == -1
        assert near_cmp(1.000001, 1) == +1

        v1 = 1000
        v2 = 1.0e-3 / 1.0e-6
        assert near_cmp(v1, v2) == 0
        assert near_cmp(v2, v1) == 0


if __name__ == "__main__":
    CompareTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_item.py $
$Id: utest_item.py 29832 2007-05-29 19:19:08Z dbinger $
"""
from datetime import datetime
from dulcinea.item import Item, ItemFolder
from qp.lib.tz import Eastern
from sancho.utest import UTest

class MyItem(Item):
    pass

class TestItem (UTest):
    def _pre(self):
        self.item1 = MyItem()
        self.item1.timestamp = datetime(2003,1,1,tzinfo=Eastern)
        self.item1.set_approved(True)
        self.item1a = MyItem()
        self.item1a.timestamp = datetime(2003,1,1,tzinfo=Eastern)
        self.item2 = MyItem()
        self.item2.timestamp = datetime(2005,1,1,tzinfo=Eastern)
        self.item2.set_approved(True)
        self.item3 = MyItem()
        self.item3.timestamp = datetime(2006,1,1,tzinfo=Eastern)
        self.item3.set_approved(True)
        self.folder = ItemFolder()

    def check_basic(self):
        self.folder.add(self.item1)
        self.folder.add(self.item2)
        assert set(self.folder.get_items()) == set([self.item1, self.item2])
        assert self.folder.get_recent_items(count=2) == [self.item2, self.item1]
        self.folder.add(self.item3)
        assert self.folder.get_recent_items(count=2) == [self.item3, self.item2]
        assert self.folder.get_recent_items(count=3) == [
            self.item3, self.item2, self.item1]
        assert self.folder.get_recent_items() == [
            self.item3, self.item2, self.item1]
        self.folder.add(self.item1a)
        assert self.folder.get_recent_items(count=2) == [
            self.item3, self.item2]
        self.item3.absorb(self.item2)
        assert self.item3.timestamp == self.item2.timestamp

    def check_delete(self):
        self.folder.add(self.item1)
        self.folder.add(self.item1a)
        item = MyItem() # auto timestamp
        self.folder.add(item)
        assert set(self.folder.get_items()) == set(
            [self.item1, self.item1a, item])
        self.folder.delete(self.item1a)
        assert set(self.folder.get_items()) == set([self.item1, item])

if __name__ == "__main__":
    TestItem()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/contact.py $
$Id: contact.py 29610 2007-03-14 21:12:59Z dbinger $
"""
from dulcinea.address import Addressable, ContactAddress
from dulcinea.history import Historical
from dulcinea.misc import get_misc_db
from dulcinea.user import DulcineaUser
from qp.lib.spec import init, add_getters_and_setters, string


class Contact (DulcineaUser, Addressable, Historical):

    prefix_is = (string, None)
    first_name_is = (string, None)
    last_name_is = (string, None)
    phone_is = (string, None)
    fax_is = (string, None)
    company_name_is = (string, None)

    def __init__(self, user_id=None):
        DulcineaUser.__init__(self, user_id=user_id)
        Addressable.__init__(self)
        Historical.__init__(self)
        init(self)

    def get_contact_address(self):
        address = self.get_address()
        return ContactAddress(
            contact_name=self.get_contact_name(),
            contact_phone_number = self.get_phone(),
            company_name=self.get_company_name(),
            **address.__getstate__())

    def get_contact_name(self, limit=35):
        last_name = self.last_name or ''
        first_name = self.first_name
        contact_name = last_name
        if first_name and len(contact_name) <= limit:
            if limit - len(contact_name) > len(first_name) + 1:
                contact_name = first_name + ' ' + last_name
            elif limit - len(contact_name) > 2:
                contact_name = first_name[0] + ' ' + last_name
        return contact_name[:limit]

    def format_realname(self):
        first_name = self.first_name or ''
        last_name = self.last_name or ''
        if first_name and last_name:
            return first_name + ' ' + last_name
        else:
            return first_name or last_name

    def format(self):
        """() -> string

        Return a string representing this user: either their real name
        (as returned by 'format_realname()') or their user ID.
        Guaranteed to return a non-empty string.
        """
        return self.format_realname() or self.id

    def record_registration (self, host):
        misc_db = get_misc_db()
        if misc_db is not None:
            misc_db.note_user_registration(self)
        self.history.add_event(None, 'register',
                               'new user ID registered from host %s' % host)

    def record_login (self, host):
        self.history.add_event(None, 'login', 'login from host %s' % host)

    def record_logout (self, host):
        self.history.add_event(None, 'logout', 'logout from host %s' % host)

    def record_failed_login (self, host):
        self.history.add_event(None, 'failed_login',
                               'failed login attempt from host %s' % host)

    def record_change (self, user, host, msg=None, password=0):
        code = password and 'change_password' or 'change'
        if msg is None:
            if password:
                msg = 'changed password'
            else:
                msg = 'changed other data'
        self.history.add_event(user, code, msg + ' (host %s)' % host)


    def get_registration_time (self):
        """() -> datetime | None
        """
        return self.history.find_last_date('register')

    def get_last_login_time (self):
        """() -> datetime | None"""
        return self.history.find_last_date('login')


add_getters_and_setters(Contact)




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/address.py $
$Id: address.py 29610 2007-03-14 21:12:59Z dbinger $
"""
from dulcinea.base import DulcineaPersistent
from dulcinea.country import get_country_codes
from qp.lib.delegation import delegate
from qp.lib.spec import either, init, specify, string, Mixin
from qp.lib.spec import spec, add_getters, add_getters_and_setters
import re


class Address (DulcineaPersistent):

    street1_is = (string, None)
    street2_is = (string, None)
    city_is = (string, None)
    state_is = (string, None)
    zip_is = (string, None)
    country_code_is = spec(
        either(None, *get_country_codes()),
        "ISO-3166 two-letter country code")

    STATES = [
        "AL", "AK", "AR", "AZ", "CA", "CO", "CT", "DE", "FL", "GA",
        "HI", "IA", "ID", "IL", "IN", "KS", "KY", "LA", "MA", "MD",
        "ME", "MI", "MN", "MO", "MS", "MT", "NC", "ND", "NE", "NH",
        "NJ", "NM", "NV", "NY", "OH", "OK", "OR", "PA", "RI", "SC",
        "SD", "TN", "TX", "UT", "VA", "VT", "WA", "WI", "WV", "WY",
        # Commonwealths & territories
        "AS", "DC", "FM", "GU", "MP", "PR", "VI",
        # Armed forces codes
        "AA", "AE", "AP"]

    def __init__(self, **kwargs):
        init(self, **kwargs)

    def is_valid(self):
        """() -> bool

        Return true if this address is valid: either it's empty or
        it's complete (and correct!).
        """
        return self._check_complete() == []

    def _check_complete(self):
        """() -> [(attrname:str, errtype:str)]

        Check whether this address is complete: must have at least one
        line of street address, city, and country; if country is US,
        must have a valid state and ZIP code too.

        Returns a list of (attrname, errtype) 2-tuples explaining any
        problems.  'errtype' is either "missing" or "bad".
        """
        errors = []
        for field in ['street1', 'city', 'country_code']:
            if not getattr(self, field):
                errors.append((field, "missing"))

        # US-specific checks
        if self.country_code == 'US':
            if not self.state:
                errors.append(('state', "missing"))
            else:
                state = self.state.upper()
                if state not in self.STATES:
                    errors.append(('state', "bad"))
                elif state != self.state:
                    self.state = state  # make uppercase

            # Check the Zip code (5 digits, or 5+4 format)
            if not self.zip:
                errors.append(('zip', "missing"))
            elif not self.has_valid_zip():
                errors.append(('zip', "bad"))

        return errors

    def has_valid_zip(self):
        if self.country_code == 'US':
            if not self.zip:
                return False
            return re.match(r'\d{5}(-\d{4})?$', self.zip) is not None
        return True

    def has_valid_state(self):
        if self.country_code == 'US':
            return self.state and self.state in self.STATES
        return True

    def error_message(self):
        msg = ""
        for errors in self._check_complete():
            msg = msg + "%s is %s, " % (
                " ".join([word.capitalize() for word in errors[0].split("_")]),
                errors[1])
        return (msg and msg[:-2] + '.') or msg

    def format(self):
        address = ""
        if self.street1:
            address += self.street1 + "\n"
        if self.street2:
            address += self.street2 + "\n"
        if self.city:
            address += self.city + ", "
        if self.state:
            address += self.state + " "
        if self.zip:
            address += self.zip
        if address[-1:] != "\n":
            address += "\n"
        if self.country_code:
            address += self.country_code
        return address

    def format_street(self):
        result = self.street1 or ''
        if self.street2:
            result += ', ' + self.street2
        return result

add_getters(Address)


class Addressable (Mixin):
    """
    Mixin for objects that contain addresses
    """
    address_is = Address

    def __init__(self, address=None, **kwargs):
        self.address = address or Address(**kwargs)

add_getters_and_setters(Addressable)


class ContactAddress (DulcineaPersistent, Addressable):
    """
    This is an address with the additional information normally needed
    for shipping.
    """
    contact_name_is = (string, None)
    contact_phone_number_is = (string, None)
    company_name_is = (string, None)

    def __init__(self, contact_name=None, contact_phone_number=None,
                 company_name=None, **kwargs):
        Addressable.__init__(self, **kwargs)
        specify(self,
                contact_name=contact_name,
                contact_phone_number=contact_phone_number,
                company_name=company_name)

    def is_valid(self):
        return self._check_complete() == []

    def error_message(self):
        return self.address.error_message()

    def set_address(self, address):
        raise RuntimeError("ContactAddress is treated as immutable")

    def has_valid_phone_number(self):
        return is_valid_phone_number(self.get_contact_phone_number(),
                                     self.get_country_code())
    def _check_complete(self):
        """() -> [(attrname:str, errtype:str)]

        Returns a list of (attrname, errtype) 2-tuples explaining any
        problems.  'errtype' is either "missing" or "bad".
        """
        errors = self.address._check_complete()
        for field in ['contact_name', 'contact_phone_number']:
            if not getattr(self, field):
                errors.append((field, "missing"))
        if self.contact_phone_number and not self.has_valid_phone_number():
            errors.append(('contact_phone_number', 'bad'))
        return errors

    def format(self):
        if self.get_contact_name():
            output = self.get_contact_name() + "\n"
        else:
            output = ""
        if self.company_name:
            output += self.company_name + "\n"
        output += self.address.format()
        return output

add_getters(ContactAddress)
delegate(
    ContactAddress,
    'address.get_street1',
    'address.get_street2',
    'address.get_city',
    'address.get_state',
    'address.get_zip',
    'address.get_country_code')


def is_valid_phone_number(number, country_code):
    if not number:
        return False
    number_of_digits = 0
    for token in number:
        if token.isdigit():
            number_of_digits += 1
    if country_code == 'US':
        return number_of_digits >= 10
    else:
        return number_of_digits >= 7

class ContactAddressable (Mixin):
    """
    Mixin for objects that have a contact address.
    """

    contact_address_is = ContactAddress

    def __init__(self):
        self.contact_address = ContactAddress()

add_getters_and_setters(ContactAddressable)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_comment.py $
$Id: utest_comment.py 31124 2008-09-16 17:35:28Z dbinger $
"""
from dulcinea.comment import Comment
from dulcinea.common import CommonTest
from dulcinea.user import DulcineaUser
from qp.pub.common import site_now

class CommentTest (CommonTest):

    def check_comment (self):
        user = DulcineaUser("jblow")
        t1 = site_now()
        c = Comment("Hello", user)
        assert c.get_text() == "Hello"
        assert c.get_user() == user
        t2 = site_now()
        assert c.text == "Hello"
        assert c.user == user
        assert t1 <= c.get_timestamp() <= t2
        assert str(c) == "%s at %s" % (user, c.get_timestamp())
        d = Comment("H", user)
        assert d != c

if __name__ == "__main__":
    CommentTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/property/test/utest_property.py $
$Id: utest_property.py 29888 2007-06-15 19:52:55Z rmasse $
"""
from dulcinea.common import CommonTest
from dulcinea.property import property_type as PT, get_property_db
from dulcinea.property.errors import PropertyTypeError
from dulcinea.property.property import Property
from dulcinea.property.property_template import MasterTemplate
from dulcinea.physical_unit import get_standard_unit
from dulcinea.physical_value import PhysicalValue
from dulcinea.tolerance import Tolerance
from sancho.utest import raises

class PropertyTest (CommonTest):

    def _pre (self):
        CommonTest._pre(self)
        if get_property_db() is None:
            return
        self.int_type = PT.get_int_property_type()
        sec = get_standard_unit('s')
        self.pv_type = PT.get_physical_value_property_type('s')
        um =  get_standard_unit('um')
        self.pv_um_type = PT.get_physical_value_property_type('um')
        self.str_list_type = PT.get_list_of_string_property_type()
        self.str_int_table_type = PT.get_string_int_table_property_type()

    def check_basics (self):
        if get_property_db() is None:
            return
        tmpl = MasterTemplate("foo", self.int_type)
        Property(tmpl)
        raises(TypeError, Property, None)
        Property(tmpl, 1)
        raises(PropertyTypeError, Property, tmpl, 1.0)
        Property(tmpl, None)
        p = Property(tmpl)
        assert p.get_name() == tmpl.get_name()
        assert p.get_type() == tmpl.get_type()
        assert p.get_title() == tmpl.get_title()
        assert p.get_name() == tmpl.get_name()
        assert p.get_type() == tmpl.get_type()
        raises(AttributeError, getattr, p, 'not')

    def check_values (self):
        if get_property_db() is None:
            return
        int_property = Property(MasterTemplate('foo', self.int_type))
        assert int_property.get_type() == self.int_type
        assert int_property.format() == None
        int_property.set_value(1)
        assert str(int_property) == "foo = 1"
        assert int_property.get_value() == 1
        int_property.set_value(2)
        assert int_property.get_value() == 2
        assert int_property.format() == "2"
        raises(PropertyTypeError, int_property.set_value, 2.1)

        pv = PhysicalValue(1.0, "s")
        tmpl = MasterTemplate("bar", self.pv_type)
        pv_property = Property(tmpl, pv)
        assert pv_property.format() == "1 s"
        raises(PropertyTypeError, pv_property.set_value, 1)
        pv2 = PhysicalValue(2.0, "s")

        list_property = Property(MasterTemplate("baz", self.str_list_type))
        list_property.set_value(['a', 'b'])
        assert list_property.format() == "a, b"

        v = {"hello": 37}
        p1 = Property(MasterTemplate('foo', self.str_int_table_type), v)
        assert p1.format() == "hello: 37"
        assert str(p1) == "foo = hello: 37"

    def check_check (self):
        if get_property_db() is None:
            return
        # Most of the work of value-checking is done by
        # MasterTemplate, so most testing is done in
        # test_template.py.  All we need to do here is ensure that
        # Property.check_value() correctly falls back on property's
        # current value if no value is passed in.
        templ = MasterTemplate('foo', self.int_type,
                               constraint=[1, 2, 3])
        property = Property(templ)
        property.set_value(3)
        raises(PropertyTypeError, property.set_value, 'foo')

    def check_copy (self):
        if get_property_db() is None:
            return
        foo_templ = MasterTemplate("foo", self.int_type,
                                   constraint=[1, 5])
        foo1 = Property(foo_templ, value=3)
        foo2 = foo1.copy()
        assert foo1 == foo2
        assert foo1.template is foo2.template
        assert foo1.value is foo2.value

        bar_templ = MasterTemplate("bar", self.pv_um_type)
        bar1 = Property(bar_templ, value=PhysicalValue(3.5, "um"))

        assert foo1 != bar1
        bar2 = bar1.copy()
        assert bar1.template is bar2.template
        assert bar1.value is bar2.value
        assert bar1.value.unit is bar2.value.unit

        foo2.set_tolerance(Tolerance(9.0))
        assert foo2 != foo1

if __name__ == "__main__":
    PropertyTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/outline.py $
$Id: outline.py 31146 2008-09-18 01:23:31Z dbinger $
"""
from durus.persistent_list import PersistentList
from qp.lib.spec import add_getters_and_setters, string_classes, either


class Outline (PersistentList):
    """
    An Outline is a list of items, which may themselves be Outlines.
    An index into an Outline may be a list of integers, and in this
    case it refers to position in one of the nested Outlines.
    """
    def iteritems(self):
        """
        Generates (int | [int], anything) in depth-first traversal order.
        """
        for j, item in enumerate(self):
            yield j, item
            if isinstance(item, Outline):
                for k, subitem in item.iteritems():
                    if isinstance(k, list):
                        yield [j] + k, subitem
                    else:
                        yield [j, k], subitem

    def iterkeys(self):
        for j, item in self.iteritems():
            yield j

    def itervalues(self):
        for item in self:
            yield item
            if isinstance(item, Outline):
                for subitem in item.itervalues():
                    yield subitem

    def index(self, obj):
        """() -> int | [int]
        Returns the first index of obj in depth-first traversal order.
        If obj is not found, raises ValueError.
        """
        for index, item in self.iteritems():
            if item == obj:
                return index
        raise ValueError

    def __getitem__(self, index):
        if isinstance(index, string_classes):
            index = self.string_to_index(index)
        if isinstance(index, (int, slice)):
            return PersistentList.__getitem__(self, index)
        else:
            next = PersistentList.__getitem__(self, index[0])
            remainder = index[1:]
            if not remainder:
                return next
            elif isinstance(next, Outline):
                return next[remainder]

    def _apply(self, method, index, *args):
        if isinstance(index, string_classes):
            index = self.string_to_index(index)
        if isinstance(index, (int, slice)):
            method(self, index, *args)
        elif len(index) == 1:
            method(self, index[0], *args)
        else:
            target = self[index[:-1]]
            if not isinstance(target, Outline):
                raise IndexError
            method(target, index[-1], *args)

    def __setitem__(self, index, value):
        self._apply(PersistentList.__setitem__, index, value)

    def __delitem__(self, index):
        self._apply(PersistentList.__delitem__, index)

    def insert(self, index, value):
        self._apply(PersistentList.insert, index, value)

    def get(self, index, default=None):
        try:
            return self[index]
        except IndexError:
            return default

    @staticmethod
    def string_to_index(str_index):
        split = str_index.split('.')
        if len(split) == 1:
            return int(split[0]) - 1
        else:
            return [int(component)-1 for component in split]

    @staticmethod
    def index_to_string(index):
        if isinstance(index, int):
            return str(index + 1)
        else:
            return '.'.join([str(component + 1) for component in index])


class TextOutline (Outline):

    text_is = either(*string_classes)

    def __init__(self, *args, **kwargs):
        Outline.__init__(self, *args, **kwargs)
        self.text = ''

add_getters_and_setters(TextOutline)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/searchable.py $
$Id: searchable.py 30312 2007-11-27 11:26:59Z dbinger $
"""
from qp.lib.spec import Mixin, unicode_string, byte_string

class Searchable (Mixin):

    search_text_is = (None, unicode_string, byte_string)

    def __init__(self):
        self.search_text = None

    def get_search_text(self):
        """() -> unicode_string | byte_string
        This method is called by searches.
        """
        if self.search_text is None:
            self.update_search_text()
        return self.search_text

    def update_search_text(self):
        """
        Compute the search text and reset the search_text attribute if there
        is any change from the current value.
        """
        new_value = str(self)
        if self.search_text != new_value:
            self.search_text = new_value





from qp.lib.spec import anything, require, sequence, string
from durus.persistent_list import PersistentList

class SpecifiedList(PersistentList):

    """
    A PersistentList that restricts the type of elements.
    Subclasses should set the class attribute `data_is` to
    be a list containing one type specifier.
    """

    data_is = [anything]

    def __init__(self, *args, **kwargs):
        PersistentList.__init__(self, *args, **kwargs)
        require(self.data, self.data_is)

    def __setitem__(self, i, item):
        if isinstance(i, int):
            require(item, self.data_is[0])
        elif isinstance(i, slice):
            require(item, sequence(self.data_is[0]))            
        PersistentList.__setitem__(self, i, item)

    def __setslice__(self, i, j, other):
        require(other, sequence(self.data_is[0]))
        PersistentList.__setslice__(self, i, j, other)

    def append(self, item):
        require(item, self.data_is[0])
        PersistentList.append(self, item)

    def insert(self, i, item):
        require(item, self.data_is[0])
        PersistentList.insert(self, i, item)

    def extend(self, other):
        require(other, sequence(self.data_is[0]))
        PersistentList.extend(self, other)


class StringList(SpecifiedList):

    """
    A PersistentList of strings.

    Instance attributes:
      data : [ string ]
    """
    data_is = [string]




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_user.py $
$Id: utest_user.py 29435 2007-02-20 20:41:24Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.user import get_realm, DulcineaUser as User, add_user
from dulcinea.user import get_matching_user
from sancho.utest import raises
from qp.pub.common import get_users

class DulcineaUserTest (CommonTest):

    def check_init (self):
        raises(TypeError, User, '!@#$%^')
        User('amk')
        u = User(None)
        u.set_id('foo')
        raises(AssertionError, User('foo').set_id, 'foo')
        raises(TypeError, User(None).set_id, '!foo')
        u = User(None)
        assert u.id == None
        u.set_id('foo')
        assert u.id == "foo"
        raises(AssertionError, u.set_id, 'bar')

    def check_password (self):
        user = User('foo')
        user.set_password("biteme")
        assert user.has_password("biteme")
        assert not user.has_password("dontbiteme")
        assert not user.has_password("biteme", realm='bogus')
        assert user.has_password("biteme", realm=get_realm())
        password1 = user.generate_password()
        assert user.has_password(password1)
        password2 = user.generate_password()
        assert user.has_password(password2)
        assert password1 != password2

    def check_add(self):
        user1 = User('1.1')
        user1.set_password('foo')
        user2 = User('2')
        add_user(user1)
        add_user(user2)
        user2.set_password('junk')
        assert get_users().get('1.1') == user1
        assert get_users()['1.1'] == user1
        raises(AssertionError, add_user, user1)

    def check_match(self):
        user1 = User('1.5')
        user1.set_email('foo@bar.com')
        add_user(user1)
        assert get_matching_user('notfound@bar.com') == None
        assert get_matching_user('foo@bar.com') == user1

if __name__ == "__main__":
    DulcineaUserTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_survey.py $
$Id: utest_survey.py 31150 2008-09-18 11:16:31Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.survey import Question, TextQuestion, SelectQuestion, Surveyable
from dulcinea.survey import QuestionDatabase, TextAnswer, SelectAnswer, Survey
from dulcinea.survey import SurveysAndQuestions, SurveysAndQuestionsDatabase
from dulcinea.user import DulcineaUser
from qp.lib.spec import Specified

class TestSurveyable (Specified, Surveyable):

    question_db_is = object

    def __init__(self, question_db=None):
        Surveyable.__init__(self)
        self.question_db = question_db

    def get_new_survey(self):
        return Survey(self.question_db.get_answers([1,2,3]))

class BadQuestion (Question):
    pass

class SurveyTest (CommonTest):

    def _pre(self):
        CommonTest._pre(self)
        self.user = DulcineaUser('u')
        self.questions = [
            TextQuestion("text1 title", sub_title="text sub title"),
            SelectQuestion("select1 title", column_headings=['Yes', 'No']),
            SelectQuestion("select2 title",
                           sub_title="check-all-that-apply question",
                           row_headings=["using machining",
                                         "using bulk ",
                                         "using substrate",
                                         "using substrate",
                                         "equipment supplier",
                                         "material supplier",
                                         "service supplier",
                                         "Other (please specify):"],
                           last_row_has_text=True),
            SelectQuestion("select3 title",
                           sub_title="This example uses a numbered radio set",
                           column_headings=list(map(str, range(1,10))),
                           row_headings=["Material type",
                                         "Thickness",
                                         "Thickness uniformity",
                                         "Resistivity",
                                         "Flatness",
                                         "Defect level"]),
            SelectQuestion("select4 title",
                           sub_title="This example uses 'range labels'",
                           range_labels = ["Not Important",
                                           "Very Important"],
                           column_headings=list(map(str, range(1,10))),
                           row_headings=["Material type",
                                         "Thickness",
                                         "Thickness uniformity",
                                         "Resistivity",
                                         "Flatness",
                                         "Defect level"])]
        self.question_db = QuestionDatabase()
        for question in self.questions:
            self.question_db.add(question)

    def check_bad_question(self):
        b1 = BadQuestion('bad')
        try:
            b1.get_new_answer()
            assert 0
        except NotImplementedError: pass

    def check_text_question(self):
         t1 = self.question_db.get(1)
         assert t1.get_new_answer().__class__ == TextAnswer
         assert t1.get_title() == "text1 title"
         assert t1.get_sub_title() == 'text sub title'

    def check_select_question(self):
         s1 = self.question_db.get(2)
         assert s1.get_new_answer().__class__ == SelectAnswer
         assert s1.get_title() == "select1 title"
         assert s1.get_sub_title() == None
         assert s1.get_column_headings() == ['Yes', 'No']
         assert s1.get_row_headings() == []
         assert s1.get_range_labels() == []
         s2 = self.question_db.get(3)
         assert s2.get_sub_title() == "check-all-that-apply question"
         assert s2.get_row_headings() == [
             "using machining",
             "using bulk ",
             "using substrate",
             "using substrate",
             "equipment supplier",
             "material supplier",
             "service supplier",
             "Other (please specify):"]
         s4 = self.question_db.get(5)
         assert s4.get_range_labels() == ["Not Important", "Very Important"]

    def check_text_answer(self):
         self.t1 = self.question_db.get(1)
         assert self.t1.get_new_answer().__class__ == TextAnswer
         a1 = self.t1.get_new_answer()
         assert a1.get_question() == self.t1
         assert a1.get_answer() == None
         assert a1.has_answer() == False
         a1.set_answer('the answer')
         assert a1.get_answer() == 'the answer'
         assert a1.has_answer() == True

    def check_select_answer(self):
         self.s1 = self.question_db.get(2)
         assert self.s1.get_new_answer().__class__ == SelectAnswer
         a2 = self.s1.get_new_answer()
         assert a2.get_question() == self.s1
         assert a2.get_answer() == None
         assert a2.has_answer() == False
         a2.set_answer(['1','2','3'])
         assert a2.get_answer() == ['1','2','3']
         assert a2.has_answer() == True
         assert a2.get_other_text() == None
         a2.set_other_text('other')
         assert a2.get_other_text() == 'other'

    def check_survey(self):
        from qp.lib.spec import get_spec_problems
        answers = self.question_db.get_answers([1,2,3,4,5])
        for j, answer in enumerate(answers):
            problems = get_spec_problems(answer)
            if problems:
                print('answer %s' % j)
                for problem in problems:
                    print(problem)

        s1 = Survey(answers)
        str(s1)
        assert len(s1) == 5
        assert len(s1.get_questions()) == 5
        assert len(s1.get_answers()) == 5
        timestamp = s1.get_timestamp()
        s1.set_timestamp()
        assert s1.get_timestamp() != timestamp
        assert s1.get_user() == None
        s1.set_user(self.user)
        assert s1.get_user().get_id() == 'u'
        assert not s1.has_answers()
        a1 = s1.get_answers()[0]
        a1.set_answer('the answer')
        assert s1.has_answers()
        s1.set_answers(s1.get_answers(), self.user)

    def check_surveyable(self):
        sv = TestSurveyable(question_db=self.question_db)
        assert sv.get_surveys() == []
        sv.add_survey(Survey(self.question_db.get_answers([1,2,3])))
        sv.get_survey(0)
        sv.get_new_survey()

    def check_surveys_and_questions(self):
        questions = [TextQuestion(
            "Do you wish to be included?",
            sub_title=("If you wish to be entered, please "
                       "enter your contact information below so you can be "
                       "notified.")),
                     SelectQuestion(
            "Are there factors more critical "
            "than hypothetical issues?", column_headings=['Yes', 'No']),
                     SelectQuestion(
            "Please choose one or more of the following categories to "
            "represent your club's activities.",
            row_headings=["maker using dirt",
                          "maker using glass",
                          "maker using metal",
                          "maker using wood",
                          "equipment supplier",
                          "material supplier",
                          "service supplier",
                          "Other (please specify):"],
            sub_title="This is a 'check-all-that-apply' style question",
            last_row_has_text=True),
                     SelectQuestion(
            "What are the key  issues?",
            sub_title="This example uses a numbered radio set",
            column_headings=list(map(str, range(1,10))),
            row_headings=["type", "ness", "uniformity",
                          "blah", "Flatness", "level"]),
                     SelectQuestion(
            "What are the key?",
            sub_title="This example uses 'range labels' in addition to numbers",
            range_labels = ["Not Important", "Very Important"],
            column_headings=list(map(str, range(1,10))),
            row_headings=["type", "ness", "uniformity",
                          "blah", "Flatness", "level"])]

        sq = SurveysAndQuestions('test', questions)
        assert sq.get_id() == 'test'
        assert len(sq.get_new_survey().get_questions()) == len(questions)
        sq_db = SurveysAndQuestionsDatabase()
        sq_db.add(sq)
        assert sq_db.get('test') is sq
        sq_db.get_mapping()['test'] is sq

if __name__ == "__main__":
    SurveyTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_timestamped.py $
$Id: utest_timestamped.py 31155 2008-09-18 15:54:09Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.timestamped import Timestamped, timestamp_sorted
from dulcinea.timestamped import reverse_timestamp_sorted
from qp.lib.spec import Specified

class Test (CommonTest):

    def a(self):
        class T(Specified, Timestamped):
            pass
        x = T()
        y = T()
        assert timestamp_sorted([x, y]) == [x,y]
        assert reverse_timestamp_sorted([x, y]) == [y,x]
        x.set_timestamp()
        assert timestamp_sorted([x, y]) == [y,x]

if __name__ == "__main__":
    Test()



#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/setup.py $
$Id: setup.py 31066 2008-09-10 14:16:21Z dbinger $

Setup script for Dulcinea
"""
from distutils import core
from distutils.extension import Extension
unused = Extension # for import checker.  Extension import is for side effect.
try:
    from qpy.setup import qpy_build_py
except ImportError:
    print("Please install Qpy first.")
    raise SystemExit
from lib.__init__ import __version__
import re, sys, os

if 'sdist' in sys.argv:
    if sys.platform == 'darwin':
        # Omit extended attributes from tarfile
        os.environ['COPYFILE_DISABLE'] = 'true'
    # Make sure that version numbers have all been updated.
    PAT = re.compile(r'\b%s\b' % re.escape(__version__))
    assert len(PAT.findall(open("LICENSE.txt").read())) == 12, __version__
    assert PAT.search(open("CHANGES.txt").readline()), __version__
    assert len(PAT.findall(open("README.txt").read())) == 1, __version__

    # Make sure that copyright statements are current.
    from datetime import datetime
    year = datetime.now().year # no tz, ok
    copyright = "Copyright (c) Corporation for National Research Initiatives %s" % year
    assert open("README.txt").read().count(copyright) == 1

kw = {'name': "Dulcinea",
      'version': __version__,
      'description': ("A package of modules useful for developing "
                      "QP/Durus applications."),
      'author': "CNRI",
      'author_email': "webmaster@mems-exchange.org",
      'url': "http://www.mems-exchange.org/software/dulcinea/",
      'license': "see LICENSE.txt",

      'package_dir': {'dulcinea':'lib'},
      'packages': ['dulcinea',
                   'dulcinea.ui',
                   'dulcinea.ui.lib',
                   'dulcinea.ui.form',
                   'dulcinea.ui.user'],
      'scripts': ['bin/expire_session.py'],
      'cmdclass': {'build_py': qpy_build_py},
     }

# If we're running Python 2.3, add extra information
if hasattr(core, 'setup_keywords'):
    if 'classifiers' in core.setup_keywords:
        kw['classifiers'] = ['Development Status :: 3 - Alpha',
          'Intended Audience :: Developers',
          'Operating System :: Unix',
          'Topic :: Software Development :: Libraries',
          ]
    if 'download_url' in core.setup_keywords:
        kw['download_url'] = ('http://www.mems-exchange.org/software/files'
                              '/dulcinea/Dulcinea-%s.tar.gz' % kw['version'])

core.setup(**kw)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/period.py $
$Id: period.py 31138 2008-09-17 02:43:20Z dbinger $
"""
from qp.lib.spec import Specified, specify, add_getters, datetime_with_tz


class Period (Specified):

    start_is = (datetime_with_tz, None)
    end_is = (datetime_with_tz, None)

    def __init__(self, start=None, end=None):
        specify(self, start=start, end=end)

    def __contains__(self, time):
        if self.start and self.start >= time:
            return False
        if self.end and self.end <= time:
            return False
        return True

    def __nonzero__(self):
        return not (self.start and self.end and self.start >= self.end)
    __bool__ = __nonzero__

    def __eq__(self, other):
        return (isinstance(other, Period) and
                self.start == other.start and self.end == other.end)

    def __str__(self):
        return 'Period(%s, %s)' % (self.start, self.end)

    def __repr__(self):
        return self.__str__()

    def is_bounded(self):
        return self.start and self.end

    def intersection(self, period):
        def start_order(a, b):
            if a is None:
                return True
            elif b is None:
                return False
            else:
                return a <= b
        def end_order(a, b):
            if b is None:
                return True
            elif a is None:
                return False
            else:
                return a <= b
        if start_order(self.start, period.start):
            start = period.start
        else:
            start = self.start
        if end_order(self.end, period.end):
            end = self.end
        else:
            end = period.end
        return Period(start, end)


add_getters(Period)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/news.py $
$Id: news.py 30485 2008-01-21 21:06:41Z dbinger $
"""
from dulcinea.base import DulcineaPersistent
from dulcinea.timestamped import Timestamped, reverse_timestamp_sorted
from qp.lib.keep import Keyed, Keep
from qp.lib.spec import specify, spec, add_getters_and_setters, string
from qp.pub.common import get_publisher

def get_news_db():
    return get_publisher().get_connection().get_root().get('news_db')

class NewsItem(Keyed, Timestamped, DulcineaPersistent):
    """
    A single announcement.
    """
    title_is = spec(
        string,
        "Title of this announcement.")
    subtitle_is = spec(
        string,
        "A subtitle that can be shown with the link text, but isn't required.")
    text_is = spec(
        string,
        "HTML text of this announcement.  This field will not be escaped.")

    def __init__ (self):
        Keyed.__init__(self)
        Timestamped.__init__(self)
        specify(self, title='', subtitle='', text='')

add_getters_and_setters(NewsItem)

class NewsDatabase(Keep, DulcineaPersistent):

    def __init__(self):
        Keep.__init__(self, value_spec=NewsItem)

    def delete_news_item(self, item):
        """(item : NewsItem)
        """
        del self.mapping[item.get_key()]

    def get_news_item(self, key):
        return self.mapping[key]

    def get_news_items(self):
        return list(self.mapping.values())

    def get_sorted_news_items(self):
        return reverse_timestamp_sorted(self.mapping.values())



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_keyed.py $
$Id: utest_keyed.py 30485 2008-01-21 21:06:41Z dbinger $
"""
from durus.persistent import Persistent
from qp.lib.keep import Keyed, Keep
from sancho.utest import UTest, raises

class Test (UTest):

    def check_a(self):
        raises(AssertionError, Keep)
        class PKeep(Keep, Persistent):
            pass
        keep = PKeep()
        class PKeyed(Keyed, Persistent):
            pass
        a = PKeyed()
        assert a.get_key() == None
        keep.add(a)
        raises(AssertionError, keep.add, a)
        b = PKeyed()
        b.junk = 1
        raises(TypeError, keep.add, b)
        raises(TypeError, keep.add, 1)
        assert a.get_key() == 1
        assert list(keep.get_mapping().items()) == [(1, a)]
        assert keep.get(1) == a

if __name__ == "__main__":
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_news.py $
$Id: utest_news.py 27635 2005-10-28 19:43:18Z dbinger $
"""
from dulcinea.news import NewsItem, NewsDatabase
from qp.lib.spec import get_spec_problems
from sancho.utest import UTest

class NewsTest(UTest):

    def test_item(self):
        a = NewsItem()
        db = NewsDatabase()
        db.add(a)
        assert a.get_title() == ""
        assert a.get_subtitle() == ""
        assert a.get_text() == ""
        assert not get_spec_problems(a), get_spec_problems(a)
        a.set_title('foo')
        assert a.get_title() == 'foo'
        a.set_subtitle('boo')
        assert a.get_subtitle() == 'boo'
        a.set_text('moo')
        assert a.get_text() == 'moo'
        a.set_timestamp()
        assert not get_spec_problems(a)
        b = NewsItem()
        db.add(b)
        assert db.get_news_item(a.get_key()) == a
        assert db.get_sorted_news_items() == [b, a]
        db.delete_news_item(b)
        assert db.get_news_items() == [a]

if __name__ == '__main__':
    NewsTest()




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/property/test/utest_property_template.py $
$Id: utest_property_template.py 28932 2006-10-12 20:29:05Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.material import Material, get_material_db
from dulcinea.property import get_property_db
from dulcinea.property.errors import PropertyTypeError, ConstraintError
from dulcinea.property.property_template import MasterTemplate
from dulcinea.property.property_type import PropertyType
from dulcinea.property.property_type import get_material_property_type
from dulcinea.property.property_type import get_physical_value_property_type
from dulcinea.property.property_type import get_string_property_type
from dulcinea.physical_unit import get_standard_unit
from dulcinea.physical_value import PhysicalValue
from dulcinea.range_value import RangeValue
from sancho.utest import raises


class TestTemplateProperty (CommonTest):

    def _pre(self):
        CommonTest._pre(self)
        if get_property_db() is None:
            return
        self.plib = get_property_db()
        get_type = self.plib.get_property_type
        self.int_type = get_type(PropertyType.ATOMIC,
                                 PropertyType.INT)
        self.material_type = get_type(PropertyType.ATOMIC,
                                      PropertyType.MATERIAL)
        pv1 = PhysicalValue(100, "degC")
        pv2 = PhysicalValue(1, "mm")
        pv3 = PhysicalValue(1, "min")
        self.pv_type = get_type(PropertyType.ATOMIC,
                                   PropertyType.PHYSICAL_VALUE,
                                   element_sub_type=pv1.unit)
        self.pv_len_type = get_type(PropertyType.ATOMIC,
                                    PropertyType.PHYSICAL_VALUE,
                                    element_sub_type=pv2.unit)
        self.pv_time_type = get_type(PropertyType.ATOMIC,
                                     PropertyType.PHYSICAL_VALUE,
                                     element_sub_type=pv3.unit)

        self.upv_type = get_type(PropertyType.ATOMIC,
                                    PropertyType.PHYSICAL_VALUE)
        self.pv_list_type = get_type(PropertyType.LIST,
                                     PropertyType.PHYSICAL_VALUE,
                                     element_sub_type=pv2.unit)
        self.pv_time_list_type = get_type(PropertyType.LIST,
                                          PropertyType.PHYSICAL_VALUE,
                                          element_sub_type=pv3.unit)
        self.mat_list_type = get_type(PropertyType.LIST,
                                      PropertyType.MATERIAL)
        self.str_pv_table_type = get_type(PropertyType.TABLE,
                                          PropertyType.PHYSICAL_VALUE,
                                          key_type=PropertyType.STRING,
                                          element_sub_type=pv2.unit)
        self.str_mat_table_type = get_type(PropertyType.TABLE,
                                           PropertyType.MATERIAL,
                                           key_type=PropertyType.STRING)


    def check_basic(self):
        if get_property_db() is None:
            return
        template = MasterTemplate("foo", self.int_type, description="bar")

        assert template.get_name() == "foo"
        assert template.get_description() == "bar"
        assert template.get_title() == "Foo"
        template.set_description(None)
        template.title=None

        template.name = "foo_bar"
        assert template.get_title() == None

        # illegal property name
        raises(TypeError, MasterTemplate, '123', self.int_type)

        # unknown type
        raises(TypeError, MasterTemplate, 'blah', 'xint')


    def check_unconstrained(self):
        if get_property_db() is None:
            return
        nocon_template = MasterTemplate('foo', self.int_type)
        assert nocon_template.get_constraint() == None
        assert nocon_template.format() == "(unconstrained)"
        assert nocon_template.explain_constraint() == None
        assert not nocon_template.is_discrete()
        nocon_template.check_value(-23432)
        nocon_template.check_value(124321431)
        raises(PropertyTypeError, nocon_template.check_value, 'foo')

    def check_simple_set_constr (self):
        if get_property_db() is None:
            return
        # material constraints are a special case of strings ...
        silicon = get_material_db().get_material('silicon')
        boron = get_material_db().get_material('boron')
        gold = get_material_db().get_material('gold')
        silicon_or_boron = Material('silicon_or_boron')
        silicon_or_boron.set_label('silicon or boron')
        silicon_or_boron.add_child(silicon)
        silicon_or_boron.add_child(boron)
        mat_template = MasterTemplate(
            'substrate', self.material_type,
            constraint=[silicon_or_boron])
        assert mat_template.get_constraint_unit() == None
        assert set(mat_template.get_constraint()) == set([silicon_or_boron])
        assert mat_template.is_discrete()
        mat_template.check_value(silicon)
        raises(ConstraintError, mat_template.check_value, gold)

        assert mat_template.explain_constraint() == "must be silicon or boron"
        assert mat_template.format() == "silicon or boron"

        int_template = MasterTemplate('foo', self.int_type,
                                      constraint=[10, 20])
        assert int_template.get_constraint_range() == RangeValue(10, 20)
        assert int_template.is_discrete()
        assert int_template.format() == "10, 20"
        assert int_template.explain_constraint() == "must be 10 or 20"

    def check_pv_set_constr (self):
        if get_property_db() is None:
            return
        # Try PhysicalValues with unit (this time, three values in the set,
        # because that's a different case in explain_constraint()).
        pv1 = PhysicalValue(100, "degC")
        pv2 = PhysicalValue(150, "degC")
        pv3 = PhysicalValue(200, "degC")
        pv4 = PhysicalValue(RangeValue(250, 300), "degC")
        pv5 = PhysicalValue(10, "degF")

        temp_template = MasterTemplate('temperature', self.pv_type,
                                       constraint=[pv1, pv2, pv3, pv4])
        assert set(temp_template.get_constraint()) == set([pv1, pv2, pv3, pv4])
        assert not temp_template.is_discrete()
        temp_template.check_value(PhysicalValue(100, 'degC'))
        temp_template.check_value(PhysicalValue(212, 'degF'))
        temp_template.check_value(PhysicalValue(270, 'degC'))
        temp_template.check_value(PhysicalValue(550, 'degF'))
        raises(ConstraintError,
               temp_template.check_value, PhysicalValue(101, 'degC'))

        assert temp_template.format(html=1) == (
                      "100, 150, 200, 250 .. 300&nbsp;&deg;C")
        assert temp_template.format(html=0) == (
                      "100, 150, 200, 250 .. 300 degC")
        assert temp_template.explain_constraint() == (
                    "must be one of 100, 150, 200, 250 .. 300 degC")
        assert temp_template.explain_constraint(html=1) == (
                    "must be one of 100, 150, 200, 250 .. 300&nbsp;&deg;C")

        # Now unitless PhysicalValues because the formatting is different.
        pv1 = PhysicalValue(50)
        pv2 = PhysicalValue(60)
        foo_template = MasterTemplate('foo', self.upv_type,
                                      constraint=[pv1, pv2])
        assert set(foo_template.get_constraint()) == set([pv1, pv2])
        foo_template.check_value(PhysicalValue(50))
        raises(PropertyTypeError, foo_template.check_value, 50)
        assert foo_template.format() == "50, 60"
        assert foo_template.explain_constraint() == "must be 50 or 60"

    def check_aggr_set_constr (self):
        if get_property_db() is None:
            return
        depth_100 = PhysicalValue(100, "um")
        depth_150 = PhysicalValue(150, "um")
        depth_200 = PhysicalValue(200, "um")

        gold = get_material_db().get_material('gold')
        aluminum = get_material_db().get_material('aluminum')
        silver = get_material_db().get_material('silver')
        boron = get_material_db().get_material('boron')
        three_metals = Material('three_metals')
        three_metals.add_child(gold)
        three_metals.add_child(aluminum)
        three_metals.add_child(silver)
        valid_depths = [depth_100, depth_200]
        matl_template = MasterTemplate("material", self.mat_list_type,
                                       constraint=[three_metals])
        depth_template = MasterTemplate("depth",
                                        self.pv_list_type,
                                        constraint=valid_depths)

        raises(PropertyTypeError, matl_template.check_value, gold)
        raises(PropertyTypeError, matl_template.check_value, boron)
        matl_template.check_value([gold])
        raises(ConstraintError, matl_template.check_value, [boron])
        raises(ConstraintError, matl_template.check_value, [gold, boron])
        matl_template.check_value([gold, silver])

        raises(PropertyTypeError, depth_template.check_value, depth_100)
        depth_template.check_value([depth_100])
        raises(ConstraintError,
               depth_template.check_value, [depth_100, depth_150])
        depth_template.check_value([depth_200, depth_100])


    def check_allowed_units(self):
        if get_property_db() is None:
            return
        min = get_standard_unit("min")
        sec = get_standard_unit("s")
        hour = get_standard_unit("hour")

        int_template = MasterTemplate('foo', self.int_type)
        assert int_template.get_default_unit() == None

        pv_template = MasterTemplate('duration', self.pv_time_type)
        assert min in pv_template.get_allowed_units()
        assert sec in pv_template.get_allowed_units()
        assert pv_template.get_default_unit() == min
        assert pv_template.is_allowed_unit(min)
        assert pv_template.is_allowed_unit('min')
        assert pv_template.is_allowed_unit(hour)

        # Ensure that check_value() enforces allowed_units
        pv1 = PhysicalValue(2, "min")
        pv2 = PhysicalValue(3, "day")
        pv3 = PhysicalValue(1.5, "m")
        pv4 = PhysicalValue(1.5)

        raises(PropertyTypeError, int_template.check_value, pv1)
        pv_template.check_value(pv1)
        raises(PropertyTypeError, pv_template.check_value, pv3)
        raises(PropertyTypeError, pv_template.check_value, pv4)
        raises(PropertyTypeError, pv_template.check_value, 1.5)

        list_template = MasterTemplate('durations',
                                       self.pv_time_list_type)
        list_template.check_value([])
        list_template.check_value([pv1])
        list_template.check_value([pv1, pv1*3])

        raises(PropertyTypeError, list_template.check_value, [pv1, pv3])
        raises(PropertyTypeError, list_template.check_value, [pv4])

    def check_consistent_units (self):
        if get_property_db() is None:
            return
        pv1 = PhysicalValue(1, "mm")
        pv2 = PhysicalValue(1.5, "mm")
        pv3 = PhysicalValue(1000, "um")
        mm_um = [pv1.unit, pv3.unit]
        templ1 = MasterTemplate('pvlist',
                                self.pv_list_type)
        templ2 = MasterTemplate('pvtable',
                                self.str_pv_table_type)

        templ1.check_value([])
        templ1.check_value([pv1])
        templ1.check_value([pv1, pv2])

        value = {'foo': pv1, 'bar': pv2}
        templ2.check_value(value)
        del value['foo']
        templ2.check_value(value)


    def check_factories (self):
        if get_property_db() is None:
            return
        templ = MasterTemplate('duration', self.pv_time_type)
        pv1 = PhysicalValue(20.0, "min")
        pv2 = PhysicalValue(25.0, "min")
        raises(PropertyTypeError, templ.create_value, 10)
        property1 = templ.create_value(pv1)
        assert property1.get_name() == templ.get_name()
        assert property1.get_type() == templ.get_type()
        assert property1.get_value() == pv1
        assert property1.template is templ

        property2 = templ.create_input_template(constraint=[pv1, pv2])
        assert property2.get_constraint() == [pv1, pv2]
        assert property2.get_name() == templ.get_name()
        assert property2.get_type() == templ.get_type()
        assert property2.template == templ

        pv3 = PhysicalValue(30.0, "min")
        assert property2.get_default_value() == None
        raises(PropertyTypeError, property2.set_default_value, 'bad')
        raises(ConstraintError, property2.set_default_value, pv3)
        property2.set_default_value(pv2)
        assert property2.get_default_value() == pv2

        set_templ = MasterTemplate("spam", self.int_type, constraint=[1, 10])
        set_templ.create_input_template(constraint=[1,10])

    def check_copy (self):
        if get_property_db() is None:
            return
        foo1 = MasterTemplate("foo", self.int_type, constraint=[1, 5])
        foo2 = foo1.copy()
        assert foo1 is not foo2
        assert foo1.name == foo2.name
        assert foo1.type is foo2.type
        assert foo1.title == foo2.title
        assert foo1.description is foo2.description is None
        assert foo1.constraint == foo2.constraint

        bar1 = MasterTemplate("bar", self.pv_len_type,
                              constraint=[PhysicalValue(1, "um"),
                              PhysicalValue(2, "um")])
        bar2 = bar1.copy()
        assert bar1 is not bar2
        assert bar1.name == bar2.name
        assert bar1.type is bar2.type
        assert bar1.constraint == bar2.constraint
        assert bar1.constraint[0] == bar2.constraint[0]
        assert bar1.constraint[0].unit is bar2.constraint[0].unit

    def check_simplify (self):
        # physical value
        if get_property_db() is None:
            return
        constraint = [PhysicalValue(RangeValue(0, 10), 'um')]
        pv_um = get_physical_value_property_type('um')
        depth_templ = MasterTemplate("depth", pv_um, constraint=constraint)
        assert depth_templ.simplify() == RangeValue(0, 10)
        depth_property = depth_templ.create_value(PhysicalValue(5300, "nm"))
        assert depth_property.simplify() == 5.3

        # material
        material_db = get_material_db()
        photoresist = material_db.get_material('gold')
        resist = material_db.get_material("gold")
        material_templ = MasterTemplate("material",
                                        get_material_property_type(),
                                        constraint=[photoresist])
        material_property = material_templ.create_value(resist)
        assert material_property.simplify() == resist.get_name()

        # string
        string_templ = MasterTemplate("letter",
                                      get_string_property_type(),
                                      constraint=["a", "b", "c"])
        string_property = string_templ.create_value("a")
        assert string_property.simplify() == "a"


if __name__ == "__main__":
    TestTemplateProperty()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/property/test/utest_property_db.py $
$Id: utest_property_db.py 28932 2006-10-12 20:29:05Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.property import get_property_db
from dulcinea.property.property_template import MasterTemplate
from dulcinea.property.property_type import get_string_property_type
from sancho.utest import raises

class PropertyDatabaseTest (CommonTest):

    def check_master_templates (self):
        if get_property_db() is None:
            return
        property_db = get_property_db()
        raises(TypeError, property_db.add_master_template, 'foo')

        s = get_string_property_type()
        self.foo = MasterTemplate('foo',s)
        self.bar = MasterTemplate('bar',s)
        raises(ValueError, property_db.rename_master_template, self.bar,'foo')

        property_db.add_master_template(self.foo)
        property_db.add_master_template(self.bar)
        raises(ValueError, property_db.add_master_template, self.foo)

        raises(ValueError, property_db.rename_master_template, self.bar, 'foo')
        property_db.rename_master_template(self.bar,'baz')
        assert self.bar == property_db.get_master_template('baz')


if __name__ == "__main__":
    PropertyDatabaseTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/physical_unit.py $
$Id: physical_unit.py 31134 2008-09-17 02:27:31Z dbinger $

Provides the PhysicalUnit and UnitCollection.
"""
from dulcinea.base import DulcineaPersistent
from durus.persistent_dict import PersistentDict
from qp.lib.spec import instance, boolean, mapping, spec, specify, require
from qp.lib.spec import string
from qp.pub.common import get_publisher
from qpy import xml, stringify


def get_standard_units():
    return get_publisher().get_connection().get_root().get('standard_units')

class UnitError (Exception):
    """Error with inconsistent or invalid physical units.
    """


class PhysicalUnit (DulcineaPersistent):
    """
    A single physical unit, consisting of a name, dimensionality,
    conversion factor to the standard (SI) unit of the same
    dimensionality, and offset to the SI unit.  (Offset is only for
    temperature units.)
    """
    name_is = spec(
        string,
        'the name of this unit in plain 7-bit ASCII (.e.g "kg*m/s^2").')
    html_name_is = spec(
        (string, None),
        "the name of the unit as it will be presented on the web.")
    powers_is = spec(
        (int, int, int, int, int, int, int),
        "denotes which SI quantity this unit measures, as a tuple of "
        "exponents on the seven fundamental SI dimensions listed in "
        "UnitCollection.DIMENSIONS.")
    units_is = spec(
        instance('UnitCollection'),
        "the UnitCollection to which this unit belongs")
    factor_is = spec(
        float,
        "the conversion factor from this unit to the standard SI unit "
        "for the same quantity (ie. with the same 'powers' tuple). "
        "By definition, the unit with factor 1.0 (and offset 0.0) is "
        "the standard SI unit for a particular quantity.")
    offset_is = spec(
        float,
        "the conversion offset from this unit to the standard SI "
        "unit for the same quantity.  Only relevant for temperature "
        "units, specifically degC and degF.")
    artificial_is = spec(
        boolean,
        "true if this unit was artificially created as a result of unit "
        "arithmetic, in which case this unit probably should not be "
        "visible to users selecting from a set of available units")


    def __init__(self, name=None, powers=None, units=None,
                 factor=1.0, offset=0.0,
                 html_name=None, artificial=0):
        self.set_name(name, html_name)
        specify(self, powers=powers, units=units, factor=factor, offset=offset,
                artificial=artificial)

    def __str__(self):
        return self.name

    def __eq__(self, other):
        return (self.__class__ == getattr(other, '__class__', None) and
            self.powers == other.powers and
            self.factor == other.factor and
            self.offset == other.offset)

    def __ne__(self, other):
        return not self == other
        
    def __hash__(self):
        return id(self)

    def set_name(self, name, html_name):
        self.name = name
        if html_name == name:
            self.html_name = None
        else:
            self.html_name = html_name

    def get_name(self, html=0):
        if html and self.html_name is not None:
            return xml(self.html_name)
        else:
            return self.name

    def get_protected_name(self, html=0):
        name = stringify(self.get_name(html=html))
        if ('/' in name or '*' in name or '^' in name):
            name = "(%s)" % name
        if html:
            return xml(name)
        else:
            return name

    def get_powers(self):
        return self.powers

    def get_conversion_factor(self, other=None):
        """(other=None : PhysicalUnit) -> factor : float

        Return the factor that converts a value in the current unit
        to units of 'other'.  'other' may be another PhysicalUnit
        instance or None (the default).  If None, just returns the
        conversion factor from the current unit to the appropriate SI
        unit, otherwise return conversion factor to the unit specified
        by 'other'.

        Raises UnitError if 'self' and 'other' are incompatible units,
        or if the conversion between them is not a simple multiplicative
        factor (eg. converting units with different offsets, such as
        degC to K or degC to degF).
        """
        if other is None:
            return self.factor
        require(other, PhysicalUnit)
        if self.powers != other.powers:
            raise UnitError("Incompatible units: %s, %s" % (self, other))
        if self.offset != other.offset and self.factor != other.factor:
            raise UnitError(
                  'Unit conversion (%s to %s) cannot be expressed ' +
                  'as a simple multiplicative factor') % \
                  (self.name, other.name)
        return self.factor/other.factor

    def get_conversion_tuple(self, other=None):
        """(other : PhysicalUnit = None) -> (factor : float, offset : float)

        Returns the (factor,offset) tuple that converts values in the
        current unit to another, compatible, unit.  Same rules for
        'other' apply as for 'get_conversion_factor()'.

        The returned conversion tuple is used as follows:
          new_value = (old_value + offset) * factor
        where old_value is in the units of 'self'; new_value will be in
        the units of 'other'.  Note that 'convert_value()' will happily
        do this arithmetic for you.

        Raises UnitError if 'self' and 'other' are incompatible units.
        """
        # no "other" unit? return tuple to convert to base unit
        if other is None:
            return (self.factor, self.offset)
        if string(other):
            other = self.units.get_unit(other)

        require(other, PhysicalUnit)

        # incompatible "other" unit? bomb!
        if self.powers != other.powers:
            raise UnitError("Incompatible units: %s, %s" % (self, other))

        # otherwise, a little arithmetic is in order... here's the
        # algebra to back it up:
        # "Let (f1,o1) be the conversion tuple from 'self' to base units
        # (ie. (x+o1)*f1 converts a value x from 'self' to base units,
        # and (x/f1)-o1 converts x from base to 'self' units)
        # and (f2,o2) be the conversion tuple from 'other' to base units
        # then we want to compute the conversion tuple (F,O) from
        #   'self' to 'other' such that (x+O)*F converts x from 'self'
        #   units to 'other' units
        # the formula to convert x from 'self' to 'other' units via the
        #   base units is (by definition of the conversion tuples):
        #     ( ((x+o1)*f1) / f2 ) - o2
        #   = ( (x+o1) * f1/f2) - o2
        #   = ( (x+o1) * f1/f2 ) - (o2*f2/f1) * f1/f2
        #   = ( (x+o1) - (o1*f2/f1) ) * f1/f2
        #   = (x + o1 - o2*f2/f1) * f1/f2
        # thus, O = o1 - o2*f2/f1 and F = f1/f2"
        factor = self.factor / other.factor
        offset = self.offset - (other.offset * other.factor / self.factor)
        return (factor, offset)

    def convert_value(self, value, other=None):
        """(value : number, other : PhysicalUnit = None) -> new_value : number

        Convert 'value', implicitly in the current unit, to a new unit
        'other'.  If 'other' is None, convert 'value' to the
        corresponding SI unit; otherwise, 'other' must be a PhysicalUnit
        object.  Raises UnitError if 'other' is incompatible with
        'self'.
        """
        require(other, (PhysicalUnit, None))
        (factor, offset) = self.get_conversion_tuple(other)
        return (value + offset) * factor

    def is_compatible(self, other):
        if other is None and self.powers is not None:
            return 0
        return self.powers == other.powers

    def is_dimension(self, name):
        """Test if the unit has the same dimensions as the supplied
        quantity name (eg. "mass", "time", etc.)
        """
        powers = self.units.quantity_powers[name]
        return self.powers == powers

    def is_inverse_compatible(self, other):
        return list(self.powers) == [-p for p in other.powers]

    def is_dimensionless(self):
        return self.powers == (0, 0, 0, 0, 0, 0, 0)

    def is_time(self):
        return self.powers == (0, 0, 1, 0, 0, 0, 0)

    def _mul(self, other):
        if self.offset != 0:
            raise UnitError(
                  "cannot multiply unit '%s' with non-zero offset %s" % 
                  (self, self.offset))

        if type(other) in (float, int):
            name = "%g*%s" % (other, self.get_protected_name())
            html_name = "%g*%s" % (other, self.get_protected_name(html=1))
            powers = self.powers
            factor = float(other) * self.factor

            return (name, html_name, powers, factor)

        elif isinstance(other, PhysicalUnit):
            if other.offset != 0:
                raise UnitError(
                      ("cannot multiply unit '%s' by other unit '%s' "
                       "with non-zero offset (%g)") %
                      (self, other, other.offset))
            name = "%s*%s" % (self.get_protected_name(),
                              other.get_protected_name())
            html_name = "%s*%s" % (self.get_protected_name(html=1),
                                   other.get_protected_name(html=1))
            powers = [p1+p2 for (p1, p2) in zip(self.powers, other.powers)]
            powers = tuple(powers)
            factor = self.factor * other.factor

            return (name, html_name, powers, factor)
        else:
            raise TypeError(
                "can only multiply physical unit by a number or "
                "another PhysicalUnit instance: not %r" % other)

    def _div(self, other):
        if self.offset != 0:
            raise UnitError(
                  "cannot divide unit '%s' with non-zero offset (%g)" % \
                  (self, self.offset))

        if type(other) in (float, int): # "self / 4"
            name = "%s/%g" % (self.name, other)
            html_name = "%s/%g" % (self.get_name(html=1), other)
            powers = self.powers
            factor = self.factor / other
            return (name, html_name, powers, factor)

        elif isinstance(other, PhysicalUnit):
            if other.offset != 0:
                raise UnitError(
                    ("cannot divide unit '%s' by unit '%s' "
                    "with non-zero offset (%g)") % (self, other, other.offset))
            name = "%s/%s" % (self.get_protected_name(),
                              other.get_protected_name())
            html_name = "%s/%s" % (self.get_protected_name(html=1),
                                   other.get_protected_name(html=1))
            powers = [p1-p2 for (p1,p2) in zip(self.powers, other.powers)]
            powers = tuple(powers)
            factor = self.factor / other.factor
            return (name, html_name, powers, factor)

        else:
            raise TypeError(
                "can only divide by a number or another PhysicalUnit instance")

    def _inv(self, junk=None):
        if self.offset != 0:
            raise UnitError(
                  "cannot invert unit '%s' with non-zero offset (%g)" % \
                  (self, self.offset))
        assert junk is None             # _arith_op() expects an 'other' arg

        name = "1/%s" % self.get_protected_name()
        if self.html_name:
            html_name = "1/%s" % self.get_protected_name(html=1)
        else:
            html_name = None
        powers = tuple([-p for p in self.powers])
        factor = 1/self.factor
        return (name, html_name, powers, factor)

    def _pow(self, n):
        if not isinstance(n, int):
            raise ValueError(
                "pow(unit, n): n must be an integer (not %r)" % n)
        if self.offset != 0:
            raise ValueError(
                "pow(unit, n): makes no sense for unit with offset")

        name = "%s^%d" % (self.get_protected_name(), n)
        html_name = "%s<sup>%d</sup>" % (self.get_protected_name (), n)
        powers = tuple([e * n for e in self.powers])
        factor = self.factor ** n
        return (name, html_name, powers, factor)


    def _arith_op(self, other, method, add=1):
        (name, html_name, powers, factor) = method(other)
        unit = self.units.reverse_lookup(powers, factor)
        if unit:
            return unit
        else:
            return self.units.create_unit(name, html_name, powers, factor,
                                          add=add, artificial=1)


    def __mul__(self, other):
        """(self : PhysicalUnit, other : number | PhysicalUnit)
           -> PhysicalUnit

        Multiply the current unit by another unit or a number and return
        a PhysicalUnit instance with appropriate dimensionality and
        conversion factor.
        """
        if other == 1:
            return self
        else:
            return self._arith_op(other, self._mul)


    def __div__(self, other):
        """(self : PhysicalUnit, other : number | PhysicalUnit)
           -> PhysicalUnit

        Divide the current unit by another unit or a number and return a
        new PhysicalUnit instance with appropriate dimensionality and
        conversion factor.  Does nothing smart about finding an existing
        PhysicalUnit for the result, since this class knows nothing
        about unit collections.
        """
        if other == 1:
            return self
        else:
            return self._arith_op(other, self._div)
    __truediv__ = __div__

    def __pow__(self, n, mod=None):
        """Raise a unit the power n (an integer).
        """
        if n == 1:
            return self
        else:
            return self._arith_op(n, self._pow)

    def inv(self):
        """Return the inverse of a unit.
        """
        return self._arith_op(None, self._inv)

    def get_artificial(self):
        return self.artificial


class UnitCollection (DulcineaPersistent):
    """
    A collection of physical units.

    Class attributes:
      dimensions : [string]
        the seven fundamental SI dimensions.  All SI quantities are
        either on this list or are arithmetically derived from members
        of this list.  Eg. "length" is an SI quantity that is also a
        fundamental dimension; "force" is a derived SI quantity:
        "mass * length / time^2".
      si_prefixes : { string : float }
        defines the SI prefixes and how much they scale a unit by
    """
    units_is = spec(
        mapping({string:PhysicalUnit}, PersistentDict),
        "the complete collection of physical units")
    unit_rev_is = spec(
        {(int,int,int,int,int,int,int):[PhysicalUnit]},
        "used for looking up units based on their quantitative "
        "description, rather than by name.  Needed by the "
        "PhysicalUnit arithmetic operator overload methods.")
    si_units_is = spec(
        {(string,None):PhysicalUnit},
        "a subset of 'units' that may have SI prefixes tacked on to "
        'create a new, valid unit.  Eg. "m" is in si_units, since "mm" '
        'and "km" are both valid units that we might want to create '
        'on-the-fly.  But "mm" is not, nor is "inch", since neither "mmm" '
        'nor "minch" make any sense.')
    quantity_powers_is = spec(
        {string:(int,int,int,int,int,int,int)},
        'defines the known SI quantities by mapping their names ("length", '
        '"force", etc.) to exponent vectors')

    DIMENSIONS = [
        'length',                       # metre
        'mass',                         # kilogram
        'time',                         # second
        'current',                      # ampere
        'temperature',                  # kelvin
        'substance_amount',             # mole
        'luminous_intensity',           # candela
        ]

    # This includes every SI prefix except "da" (*10), which is tricky
    # because it's two letters long.  Oh well, whatever.
    si_prefix = {
        'E': 1e18,                  # exa-
        'P': 1e15,                  # peta-
        'T': 1e12,                  # tera-
        'G': 1e9,                   # giga-
        'M': 1e6,                   # mega-
        'k': 1e3,                   # kilo-
        'h': 1e2,                   # hecto-
        'd': 1e-1,                  # deci-
        'c': 1e-2,                  # centi-
        'm': 1e-3,                  # milli-
        'u': 1e-6,                  # micro-
        'n': 1e-9,                  # nano-
        'p': 1e-12,                 # pico-
        'f': 1e-15,                 # femto-
        'a': 1e-18,                 # atto-
        }

    def __init__(self):
        self.units = PersistentDict()
        self.unit_rev = {}
        self.si_units = {}
        self.quantity_powers = {}

        # base dimensions
        num_dim = len(self.DIMENSIONS)
        for i, dim in enumerate(self.DIMENSIONS):
            powers = [0] * num_dim
            powers[i] = 1
            self.quantity_powers[dim] = tuple(powers)

        # First, we must define the seven base SI units, one per SI
        # dimension.  This defines the foundation of the unit system.
        self.add_base_unit("length", "m")
        self.add_base_unit("mass", "kg")
        self.add_base_unit("time", "s")
        self.add_base_unit("current", "A")
        self.add_base_unit("temperature", "K")
        self.add_base_unit("substance_amount", "mol")
        self.add_base_unit("luminous_intensity", "cd")

        # kg is the base unit of mass.
        g = PhysicalUnit("g", self["kg"].powers, self, factor=0.001)
        self._add(g)
        del self.si_units["kg"]
        self.si_units["g"] = g

        # Next, we define the SI derived units.  Unlike the calls to
        # add_base_unit(), which are based on the seven already-defined
        # SI dimensions, each of these defines a new SI "quantity" and
        # associates a canonical unit with it.  Eg. the canonical unit
        # for the quantity "area" is "m^2".

        self.add_derived_unit("area", self["m"] * self["m"], "m^2")
        self.add_derived_unit("volume", self["m^2"] * self["m"], "m^3")
        self.create_unit("Hz", None, (0, 0, -1, 0, 0, 0, 0), 1.0)
        self.add_derived_unit("frequency", self["Hz"])

        self.add_derived_unit("density", self['kg'] / self['m^3'], "kg/m^3")
        self.add_derived_unit("speed", self['m'] / self['s'], 'm/s')
        self.add_derived_unit("acceleration",
                              self['m'] / self['s']**2, "m/s^2")
        self.add_derived_unit("force",
                              self['kg'] * self['m'] / (self['s']**2), 'N')
        self.add_derived_unit("pressure", self["N"] / self['m']**2, "Pa")
        self.add_derived_unit("energy", self["N"] * self["m"], "J")
        self.add_derived_unit("power", self["J"] / self["s"], "W")
        self.add_derived_unit("intensity", self["W"] / self["m"]**2, "W/m^2")
        self.add_derived_unit("charge", self["A"] * self["s"], "C")
        self.add_derived_unit("voltage", self["W"] / self["A"], "V")
        self.add_derived_unit("resistance",
                              self["V"] / self["A"], "Ohm", "&Omega;")
        self.add_derived_unit("resistivity",
                              self["Ohm"] * self["m"], "Ohm*m", "&Omega;*m")
        self.add_derived_unit("capacitance",
                              self["A"] * self["s"] / self["V"], "F")
        self.add_derived_unit("magnetic_flux", self["V"] * self["s"], "Wb")
        self.add_derived_unit("inductance",
                              self["V"] * self["s"] / self["A"], "H")
        self.add_derived_unit("magnetic_flux_density",
                              self["Wb"] / self["m"]**2, "T")

        # Finally, we add non-SI units: each is defined in terms of an
        # existing unit, usually by just supplying a name and a
        # conversion factor.  (Sometimes we supply a conversion offset,
        # sometimes an HTML name.)  The only units we should have to
        # define here are units whose name or HTML name cannot be
        # derived from existing units, either arithmetically
        # (eg. "Ohm*cm") or through SI prefixing ("um").

        self.add_unit("Ang", self["m"], 1e-10, html_name="&Aring;")
        self.add_unit("inch", self.get_unit("cm"), 2.54)
        self.add_unit("mil", self["inch"], 0.001)

        self.add_unit("min", self["s"], 60)
        self.add_unit("hour", self["min"], 60)
        self.add_unit("day", self["hour"], 24)
        self.add_unit("week", self["day"], 7)

        # Let gf (grams of force) look like an SI unit,
        # so ugf and mgf work automatically.
        gf = self.add_unit("gf", self["N"], 0.00980665)
        self.si_units["gf"] = gf

        self.add_unit("bar", self["Pa"], 1e5)
        self.add_unit("mb", self["Pa"], 100)
        self.add_unit("atm", self.get_unit("kPa"), 101.325)
        self.add_unit("Torr", self["atm"], 1./760)
        self.add_unit("mTorr", self["Torr"], 1e-3)
        self.add_unit("psi", self["kPa"], 6.89475)
        self.add_unit("inHg", self["kPa"], 3.387)
        self.add_unit("mmHg", self["Pa"], 133.3)
        self.add_unit("degC", self["K"], 1, 273.15, html_name="&deg;C")
        self.add_unit("degF", self["degC"], 5./9., -32, html_name="&deg;F")
        self.add_unit("dyne", self["N"], 1e-5)
        self.add_unit("lb", self["N"], 4.448221615)


    def _add(self, unit):
        name = unit.name
        assert name not in self.units
        assert unit.name not in self.units
        if unit.powers in self.unit_rev:
            for u in self.unit_rev[unit.powers]:
                assert u is not unit
        self._p_note_change()
        self.units[name] = unit
        self.unit_rev.setdefault(unit.powers, []).append(unit)
        return unit

    def add_base_unit(self, quantity, name):
        assert quantity in self.DIMENSIONS, (
            "quantity %r is not a fundamental dimension" % quantity)
        powers = self.quantity_powers[quantity]
        unit = PhysicalUnit(name, powers, self)
        self.si_units[name] = unit
        return self._add(unit)

    def add_derived_unit(self, quantity, unit, name=None, html_name=None):
        if name and unit.name != name:
            unit = PhysicalUnit(name, unit.powers, units=unit.units,
                                factor=unit.factor, offset=unit.offset,
                                html_name=html_name)
            self._add(unit)
        if unit.html_name and html_name:
            unit.html_name = html_name
        self.quantity_powers[quantity] = unit.powers
        self.si_units[name] = unit

    def add_unit(self, name, base, factor=1.0, offset=0.0, html_name=None):
        (bfactor, boffset) = base.get_conversion_tuple()
        unit = PhysicalUnit(name, base.powers, self,
                            bfactor * factor,
                            boffset/factor + offset,
                            html_name=html_name)
        return self._add(unit)

    def create_unit(self, name, html_name, powers, factor,
                     add=1, artificial=0):
        unit = PhysicalUnit(name, powers, self,
                            factor=factor, html_name=html_name,
                            artificial=artificial)
        if add:
            self._add(unit)
        return unit

    def __len__(self):
        return len(self.units)

    def __getitem__(self, name):
        return self.units[name]

    def __delitem__(self, name):
        self.delete_unit(name)

    def keys(self):
        return list(self.units.keys())

    def values(self):
        return list(self.units.values())

    def has_unit(self, name):
        return name in self.units

    has_key = has_unit

    def prefix_lookup(self, name):
        if len(name) >= 2:            # prefix 1 char, unit >= 1 char
            factor = self.si_prefix.get(name[0])
            base_unit = self.si_units.get(name[1:])

            if factor and base_unit:
                if name[0] == "u":
                    html_name = "&micro;" + name[1:]
                else:
                    html_name = None
                unit = PhysicalUnit(name, base_unit.powers, self,
                                    factor=factor*base_unit.factor,
                                    html_name=html_name)
                return unit
        return None

    def reverse_lookup(self, powers, factor=1.0, offset=0.0):
        compat = self.unit_rev.get(powers)
        if compat:
            for unit in compat:
                if unit.factor == factor and unit.offset == offset:
                    return unit
        return None

    def get_unit(self, name, create=True):
        """(name : string) -> PhysicalUnit

        Lookup a unit 'name' in the collection's dictionary of known
        units.  Raise KeyError if not unit with that name exists.
        """
        if isinstance(name, PhysicalUnit):
            return name
        require(name, string)
        try:
            unit = self.units[name]
        except KeyError:
            if not create:
                raise ValueError
            unit = self.prefix_lookup(name)
            if unit is None:
                raise ValueError
            # This unit did not previously exist -- give it the name
            # preferred by the caller and store it for future reference.
            self._add(unit)
        return unit

    def get_natural_units(self):
        def get_lower_name(unit):
            return unit.get_name().lower()
        return sorted([u for u in self.units.values()], key=get_lower_name)

    def get_compatible_units(self, unit):
        """Return the list of units (as PhysicalUnit instances) that
        are compatible with 'unit' (which maybe a PhysicalUnit instance
        or a unit name (string)).
        """
        if string(unit):
            unit = self.get_unit(unit)
        require(unit, PhysicalUnit)
        return self.unit_rev[unit.powers]

    def delete_unit(self, unit):
        require(unit, PhysicalUnit)
        assert unit.name not in self.si_units, (
            "can't delete standard SI unit %r" % unit)
        assert self.units[unit.name] is unit
        assert unit in self.unit_rev[unit.powers]
        del self.units[unit.name]
        self.unit_rev[unit.powers].remove(unit)

        if unit.name in self.si_units:
            assert self.si_units[unit.name] is unit
            del self.si_units[unit.name]

    def get_conversion_factor(self, unit1, unit2=None):
        if string(unit1):
            unit1 = self.get_unit(unit1)
        require(unit1, PhysicalUnit)
        return unit1.get_conversion_factor(unit2)

    def get_conversion_tuple(self, unit1, unit2=None):
        if string(unit1):
            unit1 = self.get_unit(unit1)
        require(unit1, PhysicalUnit)
        return unit1.get_conversion_tuple(unit2)

    def convert_value(self, value, unit1, unit2=None):
        if string(unit1):
            unit1 = self.get_unit(unit1)
        require(unit1, PhysicalUnit)
        return unit1.convert_value(value, unit2)


def get_standard_unit(unit_expression):
    return get_standard_units().get_unit(unit_expression)




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_link_triple.py $
$Id: utest_link_triple.py 27889 2006-01-26 23:15:28Z rmasse $
"""

from dulcinea.link_triple import LinkTripleDatabase
from sancho.utest import UTest

class LinkTripleTest (UTest):

    def check_link_triple_db(self):
        self.ltdb = LinkTripleDatabase()
        assert len(self.ltdb.get_links()) == 0
        self.ltdb.set_links([('a', 'b', 'c'), ('1', '2', '3')])
        assert len(self.ltdb.get_links()) == 2
        assert self.ltdb[0] == ('a', 'b', 'c')
        self.ltdb.insert_link(('A', 'B', 'C'), 1)
        assert len(self.ltdb.get_links()) == 3
        assert self.ltdb.get_links() == [('a', 'b', 'c'),
                                         ('A', 'B', 'C'),
                                         ('1', '2', '3')]
        self.ltdb[2] = ('4', '5', '6')
        assert self.ltdb.get_links() == [('a', 'b', 'c'),
                                         ('A', 'B', 'C'),
                                         ('4', '5', '6')]
        del self.ltdb[0]
        assert self.ltdb.get_links() == [('A', 'B', 'C'),
                                         ('4', '5', '6')]

if __name__ == "__main__":
    LinkTripleTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/__init__.py $
$Id: __init__.py 30614 2008-03-17 15:47:42Z dbinger $
"""
__version__ = "0.18"

try:
    import dulcinea.common
except ImportError:
    print("Dulcinea %s" % __version__)



#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/bin/gc_stored_files.py $
$Id: gc_stored_files.py 30177 2007-10-19 11:30:16Z dbinger $

Delete files that are not referenced by StoredFile instances.
"""
import sys, os, glob, time
from durus.storage import gen_oid_class
from qp.lib.site import Site

# new files are not deleted since someone could be in the process of
# uploading them or attaching them in another transaction
MIN_AGE = 60*60*24

size = 0

def maybe_remove(filename):
    global size
    if (time.time() - os.stat(filename).st_mtime) > MIN_AGE:
        print(filename)
        size += os.stat(filename).st_size
        os.unlink(filename)


def main (prog, args):
    usage = "usage: %s [site]" % prog

    if len(args) != 1:
        sys.exit(usage)

    site = Site(args[0])
    connection = site.make_file_connection(readonly=True)
    used = {}
    for oid, class_name in gen_oid_class(connection.get_storage(), 'StoredFile'):
        s = connection.get(oid)
        used[s.get_full_path()] = 1
    now = time.time()
    root_directory = site.get('file_store')
    print('root_directory %s' % root_directory)
    for filename in glob.glob(os.path.join(root_directory, '?/??/*')):
        if filename not in used:
            maybe_remove(filename)
    for filename in glob.glob(os.path.join(root_directory, 'tmp/upload.*')):
        maybe_remove(filename)
    print("%s MB" % (size / (1024*1024)))

if __name__ == '__main__':
    main(sys.argv[0], sys.argv[1:])



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/metaclass.py $
$Id: metaclass.py 31351 2008-11-25 00:58:47Z dbinger $
"""
from types import FunctionType

class ClassMethodsClass (type):
    """
    A class that uses this class as its type
    gets all of its methods (except those with names that start with '__')
    converted into class methods.  This is a way to define
    classes that can act as singleton instances.

    The str method of classes with this metaclass returns the name of
    the class.
    """

    def __init__ (self, class_name, bases, namespace):
        for name, obj in namespace.items():
            if isinstance(obj, FunctionType) and not name.startswith('__'):
                setattr(self, name, classmethod(obj))

    def __str__ (self):
        return self.__name__

